#!/usr/bin/python3
#
#  Copyright (C) 2007,2020 Tom Zoerner
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# Description:
#
#  Example for the use of class Zvbi.DvbMux. This script excercises the DVB
#  multiplexer functions: The script first opens a capture device (normally
#  this will be an analog device), then continuously captures VBI data, encodes
#  it in a DVB packet stream and wites the result to STDOUT. The output stream
#  can be decoded equivalently to that of capture.py, which is:
#
#    ./dvb-mux.py | ./decode.py --pes --all

import sys
import argparse
import re
import Zvbi

# callback function invoked by Zvbi.DvbMux.feed()
def feed_cb(pkg):
    outfile.write(pkg)
    print("wrote %d" % len(pkg), file=sys.stderr)
    # return True, else multiplexing is aborted
    return True


def main_func():
    # prepare for writing binary data to STDOUT
    global outfile
    outfile = open(sys.stdout.fileno(), "wb")

    if opt.verbose:
        Zvbi.set_log_on_stderr(-1)

    opt_services = (Zvbi.VBI_SLICED_TELETEXT_B |
                    Zvbi.VBI_SLICED_VPS |
                    Zvbi.VBI_SLICED_CAPTION_625 |
                    Zvbi.VBI_SLICED_WSS_625)

    if opt.v4l2 or (opt.pid == 0 and not "dvb" in opt.device):
        opt_buf_count = 5
        opt_strict = 0
        cap = Zvbi.Capture.Analog(opt.device, services=opt_services,
                                  buffers=opt_buf_count, strict=opt_strict,
                                  trace=opt.verbose)
    else:
        cap = Zvbi.Capture.Dvb(opt.device, dvb_pid=opt.pid, trace=opt.verbose)

    if not opt.use_static:
        # create DVB multiplexer
        if not opt.raw:
            if opt.use_feed:
                mux = Zvbi.DvbMux(pes=True, callback=feed_cb)
            else:
                mux = Zvbi.DvbMux(pes=True)
            mux.set_pes_packet_size(0, 8*184)
        else:
            if opt.use_feed:
                mux = Zvbi.DvbMux(pes=True, callback=feed_cb, raw_par=cap.parameters())
            else:
                mux = Zvbi.DvbMux(pes=True, raw_par=cap.parameters())

    while True:
        # read a sliced VBI frame
        try:
            if not opt.raw:
                if opt.use_read:
                    sliced_buf = cap.read_sliced(1000)
                else:
                    sliced_buf = cap.pull_sliced(1000)
            else:
                if opt.use_read:
                    raw_buf, sliced_buf = cap.read(1000)
                else:
                    raw_buf, sliced_buf = cap.pull(1000)
                if raw_buf is None:
                    print("Capture device does not support raw capture", file=sys.stderr)
                    sys.exit(1)

            try:
                if not opt.use_static:
                    pts = int(sliced_buf.timestamp * 90000.0)
                    if not opt.raw:
                        mux.feed(opt_services, sliced_buf=sliced_buf, pts=pts)
                    else:
                        mux.feed(opt_services, sliced_buf=sliced_buf, raw_buf=raw_buf, pts=pts)

                    if not opt.use_feed:
                        for pkg in mux:
                            feed_cb(pkg)
                else: # opt.use_static
                    pkg = bytearray(2024) # multiple of 46
                    pkg_left, sliced_left = \
                        Zvbi.DvbMux.multiplex_sliced(pkg, len(pkg),
                                                     sliced_buf, len(sliced_buf),
                                                     opt_services)
                    if sliced_left != 0:
                        print("WARN: not all sliced processed", sliced_left, file=sys.stderr)
                    if pkg_left == 0:
                        print("WARN: packet buffer overflow", file=sys.stderr)
                    # output is discarded (not a valid PES stream)

            except Zvbi.DvbMuxError as e:
                print("ERROR multiplexing:", e, file=sys.stderr)

        except Zvbi.CaptureError as e:
            print("Capture error:", e)
        except Zvbi.CaptureTimeout:
            pass


def ParseCmdOptions():
    global opt
    parser = argparse.ArgumentParser(description='DVB muxing example')
    parser.add_argument("--device", type=str, default="/dev/dvb/adapter0/demux0", help="Path to video capture device")
    parser.add_argument("--pid", type=int, default=0, help="VBI channel PID for DVB")
    parser.add_argument("--v4l2", action='store_true', default=False, help="Using analog driver interface")
    parser.add_argument("--use-read", action='store_true', default=False, help="Use \"read\" capture method instead of \"pull\"")
    parser.add_argument("--use-feed", action='store_true', default=False, help="Use feed/callback API of Zvbi.DvbMux")
    parser.add_argument("--use-static", action='store_true', default=False, help="Use static packet API of Zvbi.DvbMux")
    parser.add_argument("--raw", action='store_true', default=False, help="Include raw data in output")
    parser.add_argument("--verbose", action='store_true', default=False, help="Enable trace output in the library")
    opt = parser.parse_args()

    if opt.v4l2 and (opt.pid != 0):
        print("Options --v4l2 and --pid are mutually exclusive", file=sys.stderr)
        sys.exit(1)
    if not opt.v4l2 and (opt.pid == 0) and ("dvb" in opt.device):
        print("WARNING: DVB devices require --pid parameter", file=sys.stderr)


# main
try:
    ParseCmdOptions()
    main_func()
except KeyboardInterrupt:
    sys.stdout.flush()



#!/usr/bin/python3
#
#  libzvbi test of raw data capturing and raw decoding
#
#  Copyright (C) 2000-2002, 2004 Michael H. Schimek
#  Copyright (C) 2003 James Mastros
#  Perl Port: Copyright (C) 2007 Tom Zoerner
#  Python Port: Copyright (C) 2020 Tom Zoerner
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# Description:
#
#   Example for the use of class Zvbi.RawDec. The script continuously
#   captures raw VBI data and displays the data as an animated gray-scale
#   image. Below this, the analog wave line of one selected video line is
#   plotted (i.e. essentially simulating an oscilloscope). For the
#   selected line, the resulting data from slicing is also shown if
#   decoding is successful.
#
#   (This script is loosely based on test/osc.c in the libzvbi package.)

import sys
import re
import argparse
from tkinter import *
import Zvbi

cap = None
par = None
rawdec = None
src_w = -1
src_h = -1
dst_h = 64
raw1 = None

tk = None
canvas = None
pgm = None
canvas_pgm = None
canvas_lid = None
dec_text = None

draw_row = None
draw_count = None

# -----------------------------------------------------------------------------
# from capture.c

def PIL(day, mon, hour, minute):
    return ((day << 15) + (mon << 11) + (hour << 6) + (minute << 0))


def decode_ttx(buf):
    packet_address = Zvbi.unham16p(buf)
    if packet_address < 0:  # hamming error
        return ""

    magazine = packet_address & 7
    packet = packet_address >> 3

    text = ("pg %x%02d >" % (magazine, packet))

    astr = Zvbi.unpar_str(buf, '.')
    astr = re.sub(r'[\x00-\x1F\x7F]', '.', astr.decode('ISO-8859-1'))

    return text + astr[0 : 42] + "<"


def dump_pil(pil):
    day = pil >> 15
    mon = (pil >> 11) & 0xF
    hour = (pil >> 6) & 0x1F
    minute = pil & 0x3F

    if pil == PIL(0, 15, 31, 63):
        text = " PDC: Timer-control (no PDC)"
    elif pil == PIL(0, 15, 30, 63):
        text = " PDC: Recording inhibit/terminate"
    elif pil == PIL(0, 15, 29, 63):
        text = " PDC: Interruption"
    elif pil == PIL(0, 15, 28, 63):
        text = " PDC: Continue"
    elif pil == PIL(31, 15, 31, 63):
        text = " PDC: No time"
    else:
        text = (" PDC: %05x, 200X-%02d-%02d %02d:%02d" %
                (pil, mon, day, hour, minute))

    return text


pr_label = ""
tmp_label = ""

def decode_vps(buf):
    global pr_label
    global tmp_label

    if not opt.dump_vps:
        return

    text = "VPS: "

    c = Zvbi.rev8(buf[1])

    if (c & 0x80):
        pr_label = tmp_label
        tmp_label = ""

    c &= 0x7F
    if (c >= 0x20) and (c < 0x7f):
        tmp_label += chr(c)
    else:
        tmp_label += '.'

    text += (" 3-10: %02x %02x %02x %02x %02x %02x %02x %02x (\"%s\")" %
             (buf[0], buf[1], buf[2], buf[3], buf[4], buf[5], buf[6], buf[7], pr_label))

    pcs = buf[2] >> 6

    cni = (  ((buf[10] & 3) << 10)
           + ((buf[11] & 0xC0) << 2)
           + ((buf[8] & 0xC0) << 0)
           + (buf[11] & 0x3F))

    pil = ((buf[8] & 0x3F) << 14) + (buf[9] << 6) + (buf[10] >> 2)

    pty = buf[12]

    text += (" CNI: %04x PCS: %d PTY: %d " % (cni, pcs, pty))

    text += dump_pil(pil)

    return text

# End from capture.c
# -----------------------------------------------------------------------------

def draw(v=None):
    global raw1

    if v:
        # new frame available - handling freeze & single-stepping counter
        if draw_count.get() == 0:
            return
        if draw_count.get() > 0:
            draw_count.set(draw_count.get() - 1)

        # store a copy of the raw data (to allow navigating during freeze)
        raw1 = v

    if not raw1:
        return

    # display raw data as gray-scale image
    draw_pgm()

    if (draw_row.get() >= 0) and (draw_row.get() < src_h):
        field = (draw_row.get() >= par.count_a)

        if (par.start_a if (field == 0) else par.start_b) < 0:
            nchars = ("Row %d Line ? - " % draw_row.get())
            phys_line = -1

        elif field == 0:
            phys_line = draw_row.get() + par.start_a
            nchars = ("Row %d Line %d - " % (draw_row.get(), phys_line))

        else:
            phys_line = draw_row.get() - par.count_a + par.start_b
            nchars = ("Row %d Line %d - " % (draw_row.get(), phys_line))

        sliced_buf = rawdec.decode(raw1)

        # search the selected physical line in the slicer output
        for sld, slid, slin in sliced_buf:
            if slin == phys_line:
                # display decoder output
                draw_dec(nchars, sld, slid, slin)
                break
        else:
            # print "unkown"
            draw_dec(nchars)

        # plot that line as a wave line
        draw_plot()


def draw_pgm():
    global pgm

    pgm_str = bytearray("P5\n%d %d\n255\n" % (src_w, src_h), 'ascii') + raw1

    # image data is accepted only as bytes object, not bytearray
    pgm_str = bytes(pgm_str)

    new_pgm = tk.call("image", "create", "photo", "-data", pgm_str)

    canvas.itemconfigure(canvas_pgm, image=new_pgm)

    tk.call("image", "delete", pgm)
    pgm = new_pgm


def draw_dec(nchars, sld=None, slid=None, slin=None):
    if sld:
        if (slid & Zvbi.VBI_SLICED_TELETEXT_B):
            nchars += decode_ttx(sld)

        elif (slid & Zvbi.VBI_SLICED_VPS):
            nchars += decode_vps(sld)

        elif (slid & (Zvbi.VBI_SLICED_CAPTION_625 |
                      Zvbi.VBI_SLICED_CAPTION_525)):
            nchars += "Closed Caption"

        else:
            nchars += ("Sliced service 0x%X" % slid)
    else:
        nchars += "Unknown signal"

    dec_text.set(nchars)


def draw_plot():
    start = src_w * draw_row.get()

    Poly = []
    r = src_h + 0 + dst_h
    for i in range(0, src_w):
        y = raw1[start + i]
        Poly.append(i)
        Poly.append(r - y * dst_h//256)

    canvas.coords(canvas_lid, Poly)


def change_row(step):
    row = draw_row.get()
    if (step > 0):
        if (row + 1 < src_h):
            row += 1
    elif (step < 0):
        if (row > 0):
            row -= 1
    draw_row.set(row)

    draw()
    return "break"


def change_count(cnt):
    draw_count.set(cnt)


def resize_window(w, h):
    global canvas_lid
    global dst_h

    dst_h = h - (src_h + 10)
    if dst_h < src_h:
        dst_h = src_h

    # remove old grid
    for cid in canvas.find('overlapping', 0, src_h + 1,
                           src_w, src_h + 1 + dst_h*2):
        canvas.delete(cid)

    # paint grid in the new size
    x = 0
    while x < src_w:
       canvas.create_line(x, src_h + 10, x, src_h + 12+dst_h, fill='#AAAAAA')
       x += 10

    # create plot element
    canvas_lid = canvas.create_line(0, src_h+dst_h, 0, src_h+dst_h, fill='#ffffff')

    if draw_count == 0:
        draw(None)


def init_window():
    global tk
    global canvas
    global pgm
    global canvas_pgm
    global dec_text
    global draw_row
    global draw_count

    tk = Tk(className="osc")
    tk.wm_title('Raw capture & plot')

    dec_text = StringVar(tk, "")
    draw_row = IntVar(tk, 0)
    draw_count = IntVar(tk, -1)

    canvas = Canvas(tk, borderwidth=1, relief=SUNKEN, background='#000000',
                          height=src_h + 10 + dst_h, width=640,
                          scrollregion=[0, 0, src_w, src_h])
    canvas.pack(side=TOP, fill=X, expand=True)
    csb = Scrollbar(tk, orient=HORIZONTAL, takefocus=False, width=10, borderwidth=1,
                        command=canvas.xview)
    canvas.configure(xscrollcommand=csb.set)
    csb.pack(side=TOP, fill=X)
    canvas.pack(side=TOP, fill=BOTH, expand=True)

    canvas.bind('<Configure>', lambda e: resize_window(e.width, e.height))
    canvas.bind('<Key-q>', lambda e: sys.exit(0))
    canvas.bind('<Down>', lambda e: change_row(1))
    canvas.bind('<Up>', lambda e: change_row(-1))
    canvas.bind('<space>', lambda e: change_count(1))  # single-stepping
    canvas.bind('<Return>', lambda e: change_count(-1))  # live capture
    canvas.bindtags([canvas, 'all'])  # remove widget default bindings
    canvas.focus_set()

    label = Entry(tk, textvariable=dec_text, takefocus=0, width=50)
    label.pack(side=TOP, fill=X, anchor=W)

    f = Frame(tk)
    f_c = Checkbutton(f, text='Live capture',
                         offvalue=1, onvalue=-1, variable=draw_count)
    f_l = Label(f, text="Plot row:")
    f_s = Spinbox(f, from_=0, to=src_h - 1, width=5,
                     textvariable=draw_row, command=draw)
    f_c.pack(side=LEFT, padx=10, pady=5)
    f_l.pack(side=LEFT, padx=5)
    f_s.pack(side=LEFT, padx=5, pady=5)
    f.pack(side=TOP, anchor=W)

    pgm = tk.call("image", "create", "photo")
    canvas_pgm = canvas.create_image(0, 0, image=pgm, anchor=NW)


def cap_frame():
    # note: must use "read" and not "pull" since a copy of the data is kept in "raw1"
    try:
        raw_buf = cap.read_raw(10)
        draw(raw_buf)

    except Zvbi.CaptureError as e:
        if not opt.ignore_error:
            print("Capture error:", e, file=sys.stderr)
    except Zvbi.CaptureTimeout:
        # not an error/warning as this occurs regularly due to polling via 10ms timer
        pass

    tk.after(10, cap_frame)


def main_func():
    global cap
    global rawdec
    global par
    global src_w
    global src_h

    if opt.verbose > 1:
        vbi_capture_set_log_fp(cap, stderr)
        Zvbi.set_log_on_stderr(0)

    services = ( Zvbi.VBI_SLICED_VBI_525 |
                 Zvbi.VBI_SLICED_VBI_625 |
                 Zvbi.VBI_SLICED_TELETEXT_B |
                 Zvbi.VBI_SLICED_CAPTION_525 |
                 Zvbi.VBI_SLICED_CAPTION_625 |
                 Zvbi.VBI_SLICED_VPS |
                 Zvbi.VBI_SLICED_WSS_625 |
                 Zvbi.VBI_SLICED_WSS_CPR1204 )

    opt_strict = 0
    opt_buf_count = 5
    opt_scanning = (525 if opt.ntsc else (625 if opt.pal else 0))

    if opt.v4l2 or (opt.pid == 0 and not "dvb" in opt.device):
        cap = Zvbi.Capture.Analog(opt.device, services=services, scanning=opt_scanning,
                                  buffers=opt_buf_count, strict=opt_strict, trace=opt.verbose)
    else:
        cap = Zvbi.Capture.Dvb(opt.device, dvb_pid=opt.pid, trace=opt.verbose)

    rawdec = Zvbi.RawDec(cap)
    rawdec.add_services(services, opt_strict)

    par = cap.parameters()
    if not par.sampling_format == Zvbi.VBI_PIXFMT_YUV420:
        print(("Unexpected sampling format:%d\n" +
               "In raw decoder parameters: %s\n" +
               "Likely the device does not support capturing raw data")
                % (par.sampling_format, str(par)), file=sys.stderr)
        sys.exit(1)

    src_w = par.bytes_per_line
    src_h = par.count_a + par.count_b

    init_window()

    # install a Tk event handler for capturing in the background
    #io = new IO::Handle
    #io.fdopen(cap.get_fd(), 'r')
    #tk.fileevent(io, 'readable', cap_frame)
    tk.after(10, cap_frame)

    # everything from here on is event driven
    tk.mainloop()


def ParseCmdOptions():
    global opt
    parser = argparse.ArgumentParser(description="Plot raw captured VBI video lines")
    parser.add_argument("--device", type=str, default="/dev/vbi0", help="Path to video capture device")
    parser.add_argument("--pid", type=int, default=0, help="Teletext channel PID for DVB")
    parser.add_argument("--v4l2", action='store_true', default=False, help="Using analog driver interface")
    parser.add_argument("--pal", action='store_true', default=False, help="Force PAL video format")
    parser.add_argument("--ntsc", action='store_true', default=False, help="Force NTSC video format")
    parser.add_argument("--ignore-error", action='store_true', default=False, help="Ignore errors silently")
    parser.add_argument("--verbose", action='count', default=0, help="Enable trace in VBI library")
    opt = parser.parse_args()

    if opt.v4l2 and (opt.pid != 0):
        print("Options --v4l2 and --pid are mutually exclusive", file=sys.stderr)
        sys.exit(1)
    if not opt.v4l2 and (opt.pid == 0) and ("dvb" in opt.device):
        print("WARNING: DVB devices require --pid parameter", file=sys.stderr)


ParseCmdOptions()
main_func()



#!/usr/bin/python3
#
#  libzvbi test
#
#  Copyright (C) 2003, 2005 Michael H. Schimek
#  Perl Port: Copyright (C) 2007 Tom Zoerner
#  Python Port: Copyright (C) 2020 Tom Zoerner
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# Description:
#
#   Automated test of the odd parity and Hamming encoder and decoder
#   functions. Note this test runs for a long time.
#
#   (This is a direct translation of test/hamm.c in the libzvbi package.)

import sys
import random
import Zvbi

def sizeof(n):
    return 4

def CHAR_BIT():
    return 8

def cast_unsigned_int(v):
    return v & 0xFFFFFFFF

def cast_int(v):
    return v & 0xFFFFFFFF

def mrand48():
    val = random.randrange(0xFFFFFFF);
    if random.randrange(2) == 0:
        val *= -1
    return val

def parity(n):
    sh = sizeof(n) * CHAR_BIT() // 2;
    while sh > 0:
        n ^= n >> sh
        sh >>= 1

    return n & 1

def BC(v):
    return (v * cast_unsigned_int(0x01010101))  #0x0101010101010101

def population_count(n):
    n -= (n >> 1) & BC (0x55)
    n = (n & BC (0x33)) + ((n >> 2) & BC (0x33))
    n = (n + (n >> 4)) & BC (0x0F)

    return (n * BC (0x01)) >> (sizeof ("unsigned int") * 8 - 8)


def hamming_distance(a, b):
    return population_count(a ^ b)

def main_func():
    print("Testing parity...")

    for i in range (0, 10000):
        n = i if (i < 256) else cast_unsigned_int(mrand48 ())
        buf = bytes([n&0xFF, (n >> 8)&0xFF, (n >> 16)&0xFF])

        r = 0
        for j in range(0, 8):
            if (n & (0x01 << j)):
                r |= 0x80 >> j

        assert (r == Zvbi.rev8(n))

        if (parity (n & 0xFF)):
            assert (Zvbi.unpar8(n & 0xFF) == cast_int(n & 127))
        else:
            assert (-1 == Zvbi.unpar8(n & 0xFF))

        assert (Zvbi.unpar8(Zvbi.par8(n & 0x7F)) >= 0)

        buf2 = Zvbi.par_str(buf)
        buf3 = Zvbi.unpar_str(buf2)
        assert (0 == ((buf3[0] | buf3[1] | buf3[2]) & 0x80))

        tmp1 = Zvbi.par8(buf3[1])
        tmp2 = tmp1 ^ 0x80
        buf4 = bytes([0, tmp1, tmp2])

        buf5 = Zvbi.unpar_str(buf4)
        assert (buf5[2] == (buf5[1] & 0x7F))

    print("OK")
    print("Testing Hamming-8/4...")

    for i in range(0, 10000):
        n = i if (i < 256) else cast_unsigned_int(mrand48())

        A = parity(n & 0xA3)
        B = parity(n & 0x8E)
        C = parity(n & 0x3A)
        D = parity(n & 0xFF)

        d = (  ((n & 0x02) >> 1)
             + ((n & 0x08) >> 2)
             + ((n & 0x20) >> 3)
             + ((n & 0x80) >> 4))

        if (A and B and C):
            nn = n if D else (n ^ 0x40)

            assert (Zvbi.ham8(d) == (nn & 255))
            assert (Zvbi.unham8(nn) == d)

        elif not D:
            dd = Zvbi.unham8(n)
            assert (dd >= 0 and dd <= 15)

            nn = Zvbi.ham8(dd)
            assert (hamming_distance(n & 255, nn) == 1)

        else:
            assert (Zvbi.unham8(n) == -1)

        #Zvbi.ham16 (buf, n)
        #assert (Zvbi.unham16 (buf) == (int)(n & 255))

    print("OK")
    print("Testing Hamming-24/18...")

    for i in range(0, (1 << 24)):
        buf = bytes([i&0xFF, (i >> 8)&0xFF, (i >> 16)&0xFF])

        A = parity (i & 0x555555)
        B = parity (i & 0x666666)
        C = parity (i & 0x787878)
        D = parity (i & 0x007F80)
        E = parity (i & 0x7F8000)
        F = parity (i & 0xFFFFFF)

        d = (  ((i & 0x000004) >> (3 - 1))
             + ((i & 0x000070) >> (5 - 2))
             + ((i & 0x007F00) >> (9 - 5))
             + ((i & 0x7F0000) >> (17 - 12)))
        
        if (A and B and C and D and E):
            assert (Zvbi.unham24p(buf) == d)
        elif F:
            assert (Zvbi.unham24p(buf) < 0)
        else:
            err = ((E << 4) | (D << 3)
                     | (C << 2) | (B << 1) | A) ^ 0x1F

            assert (err > 0)

            if (err >= 24):
                assert (Zvbi.unham24p(buf) < 0)
                continue

            ii = i ^ (1 << (err - 1))

            A = parity (ii & 0x555555)
            B = parity (ii & 0x666666)
            C = parity (ii & 0x787878)
            D = parity (ii & 0x007F80)
            E = parity (ii & 0x7F8000)
            F = parity (ii & 0xFFFFFF)

            assert (A and B and C and D and E and F)

            d = (+ ((ii & 0x000004) >> (3 - 1))
                  + ((ii & 0x000070) >> (5 - 2))
                  + ((ii & 0x007F00) >> (9 - 5))
                  + ((ii & 0x7F0000) >> (17 - 12)))

            assert (Zvbi.unham24p(buf) == d)

        if (i & 0x00FFFF) == 0:
            print(".", end='')
            sys.stdout.flush()

    print("OK")

main_func()



#!/usr/bin/python3
#
#  libzvbi test of teletext search
#
#  Copyright (C) 2007,2020 Tom Zoerner
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

# Description:
#
#   Example for the use of class Zvbi.Search. The script captures and
#   caches teletext pages until the RETURN key is pressed, then prompts
#   for a search string. A search across teletext pages is started, and
#   the content of matching pages is printed on the terminal.

import os
import sys
import select
import argparse
import threading
import Zvbi

cr = '\r' if sys.stdout.isatty() else "\n"
last_pgno = -1
last_subno = -1
page_cache = dict()

def progress(pg, search_cnt):
    (page, sub) = pg.get_page_no()
    print("%sSearching page %03x.%02X " % (cr, page, sub),
          file=sys.stderr, end='')
    search_cnt[0] += 1
    return True  # let search continue


def search(vtdec):
    global last_pgno
    global last_subno

    match_cnt = 0
    search_cnt = [0]

    pat = input("\nEnter search pattern: ")
    if pat:
        #rand = [int(rand(1000))]
        #print "Search rand user data rand->[0]\n"
        srch = Zvbi.Search(vtdec, pat, casefold=False, regexp=False,
                           progress=progress, userdata=search_cnt)

        for pg in srch:
            (pgno, subno) = pg.get_page_no()
            if (last_pgno < 0) or (pgno != last_pgno) or (subno != last_subno):
                print("Found match: %03X.%04X\n" % (pgno, subno))
                print(pg.print_page(fmt='ascii').decode('utf-8'))
                last_pgno = pgno
                last_subno = subno
                match_cnt += 1

        print("%sFound %d matches on %d searched pages for \"%s\""
                % (cr, match_cnt, search_cnt[0], pat));


def pg_handler(page, subno, pg_cnt):
    print("%sCached %d pages (last: %03x.%02x) " %
            (cr, pg_cnt, page, subno), file=sys.stderr, end='')


# event handler registered with service decoder - invoked in capture thread
def capture_handler(pgtype, ev):
    global thr_page
    if pgtype == Zvbi.VBI_EVENT_TTX_PAGE:
        pgsub = (ev.pgno << 16) | ev.subno;
        # provide event data to main thread via shared memory
        with thr_lock:
            thr_page = pgsub
            page_cache[pgsub] = True;
        # dummy write to pipe is only used for waking up main thread
        os.write(thr_pipe[1], bytes([0]));


# capture loop, started in separate thread
# (i.e. running concurrently even during search pattern input & search)
def capture_loop():
    while True:
        try:
            sliced_buf = cap.pull_sliced(1000)
            vtdec.decode(sliced_buf)
        except Zvbi.CaptureError as e:
            print("Capture error:", e, file=sys.stderr)
        except Zvbi.CaptureTimeout:
            print("Capture timeout:", file=sys.stderr)


def main_func():
    global cap
    global vtdec
    global thr_inst
    global thr_pipe
    global thr_lock

    if opt.v4l2 or (opt.pid == 0 and not "dvb" in opt.device):
        opt_buf_count = 5
        opt_strict = 0
        cap = Zvbi.Capture.Analog(opt.device, services=services,
                                  buffers=opt_buf_count, strict=opt_strict, trace=opt.verbose)
    else:
        cap = Zvbi.Capture.Dvb(opt.device, dvb_pid=opt.pid, trace=opt.verbose)

    vtdec = Zvbi.ServiceDec()
    vtdec.event_handler_register(Zvbi.VBI_EVENT_TTX_PAGE, capture_handler)

    # instantiate capture thread & IPC
    thr_pipe = os.pipe2(os.O_NONBLOCK)
    thr_lock = threading.Lock()
    thr_inst = threading.Thread(target=capture_loop, daemon=True)
    thr_inst.start()

    print("Press RETURN for entering a search pattern", file=sys.stderr)

    while True:
        # check if STDIN is readable (i.e. RETURN key was pressed)
        ret = select.select([sys.stdin.fileno(), thr_pipe[0]], [], [], 0)[0]
        if sys.stdin.fileno() in ret:
            # discard one line of input (i.e. empty line caused by RETURN key)
            input()
            # ask for search string & perform search
            search(vtdec)
            print("\nPress RETURN for entering a search pattern", file=sys.stderr)

        elif thr_pipe[0] in ret:
            # drain pipe (content is unused)
            os.read(thr_pipe[0], 1000)
            with thr_lock:
                page = thr_page >> 16
                subno = thr_page & 0xFFFF
                pg_cnt = len(page_cache)
            pg_handler(page, subno, pg_cnt)


def ParseCmdOptions():
    global opt
    parser = argparse.ArgumentParser(description='Example for searching through teletext page content')
    parser.add_argument("--device", type=str, default="/dev/dvb/adapter0/demux0", help="Path to video capture device")
    parser.add_argument("--pid", type=int, default=0, help="Teletext channel PID for DVB")
    parser.add_argument("--v4l2", action='store_true', default=False, help="Using analog driver interface")
    parser.add_argument("--verbose", action='store_true', default=False, help="Enable trace output in the library")
    opt = parser.parse_args()

    if opt.v4l2 and (opt.pid != 0):
        print("Options --v4l2 and --pid are mutually exclusive", file=sys.stderr)
        sys.exit(1)
    if not opt.v4l2 and (opt.pid == 0) and ("dvb" in opt.device):
        print("WARNING: DVB devices require --pid parameter", file=sys.stderr)


try:
    ParseCmdOptions()
    main_func()
except KeyboardInterrupt:
    print("")  # needed due to progress output without trailing NL



#!/usr/bin/python3
#
#  libzvbi test of export interfaces
#
#  Copyright (C) 2000, 2001 Michael H. Schimek
#  Perl Port: Copyright (C) 2007 Tom Zoerner
#  Python Port: Copyright (C) 2020 Tom Zoerner
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# Description:
#
#   Example for the use of export option management in class Zvbi.Export.
#   Test of page export options and menu interfaces. The script lists all
#   available export modules (i.e. formats) and options.
#
#   (This is a direct translation of test/explist.c in the libzvbi package.)

import sys
import re
import argparse
import Zvbi

TypeStr = {Zvbi.VBI_OPTION_BOOL: "VBI_OPTION_BOOL",
           Zvbi.VBI_OPTION_INT: "VBI_OPTION_INT",
           Zvbi.VBI_OPTION_REAL: "VBI_OPTION_REAL",
           Zvbi.VBI_OPTION_STRING: "VBI_OPTION_STRING",
           Zvbi.VBI_OPTION_MENU: "VBI_OPTION_MENU"}

def INT_TYPE(v):
    return ((v['type'] == Zvbi.VBI_OPTION_BOOL) or
            (v['type'] == Zvbi.VBI_OPTION_INT) or
            (v['type'] == Zvbi.VBI_OPTION_MENU))

def REAL_TYPE(v):
    return (v['type'] == Zvbi.VBI_OPTION_REAL)

def MENU_TYPE(v):
    return not (v.get('menu') is None)

def BOUNDS_CHECK(oi):
    if MENU_TYPE(oi):
        if ((oi['def'] >= 0) or
            (oi['def'] <= oi['max']) or
            (oi['min'] == 0) or
            (oi['max'] > 0) or
            (oi['step'] == 1)):
            pass
        else:
            raise Exception("ERROR: bounds check for menu failed:", str(oi));
    else:
        if ((oi['max'] >= oi['min']) or
            (oi['step'] > 0) or
            ((oi['def'] >= oi['min']) and (oi['def'] <= oi['max']))):
            pass
        else:
            raise Exception("ERROR: bounds check for menu failed:", str(oi));


def STRING_CHECK(oi):
    if MENU_TYPE(oi):
        if ((oi['def'] >= 0) or
            (oi['def'] <= oi['max']) or
            (oi['min'] == 0) or
            (oi['max'] > 0) or
            (oi['step'] == 1)):
            pass
        else:
            raise Exception("ERROR: check for STRING failed:", str(oi));
    else:
        if oi.get('def') is None:
            raise Exception("ERROR: check for STRING failed:", str(oi));

def keyword_check(keyword):
    if ((keyword is None) or not re.match(r'^[a-zA-Z0-9_]+', keyword)):
        raise Exception("ERROR: invalid keyword: \"%s\"" % keyword);

def print_current(oi, current):
    if REAL_TYPE(oi):
        print("    current value=%f" % current)
        if MENU_TYPE(oi):
            if (current < oi['min']) or (current > oi['max']):
                raise Exception("ERROR: current:%d not in range %d..%d" % (current, oi['min'], oi['max']));
    else:
        print("    current value=%d" % current)
        if not MENU_TYPE(oi):
            if (current < oi['min']) or (current > oi['max']):
                raise Exception("ERROR: current:%d not in range %d..%d" % (current, oi['min'], oi['max']));

def test_modified(oi, old, new):
    #if not old == new:
    #    print("but modified current value to %s" % new)
    pass

def test_set_int(ex, oi, current, value):
    print("    trying to set %d: " % value, end='')

    try:
        ex.option_set(oi['keyword'], value)
        print("success.")
        test_modified(oi, current, new_current)
    except Zvbi.ExportError as e:
        print("ERROR: export failed:", e)

    try:
        #new_current = 0x54321
        new_current = ex.option_get(oi['keyword'])
        test_modified(oi, current, new_current)
    except Zvbi.ExportError as e:
        print("option_get failed:", e)
        #if not new_current == 0x54321:
        #    print("but modified destination to new_current")

    current = new_current
    print_current(oi, new_current)


def test_set_real(ex, oi, current, value):
    print("    try to set %f: " % value)
    try:
        ex.option_set(oi['keyword'], value)
        print("success.")
    except Zvbi.ExportError as e:
        print("option_get failed:", e)

    #new_current = 8192.0
    try:
        new_current = ex.option_get(oi['keyword'])
        test_modified(oi, current, new_current)
    except Zvbi.ExportError as e:
            print("option_get failed", e)
            # XXX unsafe
            #if not new_current == 8192.0:
            #    print("but modified destination to new_current")

    current = new_current
    print_current(oi, new_current)


def test_set_entry(ex, oi, current, entry):
    valid = (MENU_TYPE(oi)
                 and (entry >= oi['min'])
                 and (entry <= oi['max']))

    print("    try to set menu entry %d: " % entry, end='')

    try:
        ex.option_menu_set(oi['keyword'], entry)
        if valid:
            print("success.")
        else:
            print("ERROR: option_menu_set unexpected success.")

        new_current = ex.option_get(oi['keyword'])
        test_modified(oi, current, new_current)

    except Zvbi.ExportError as e:
        if valid:
            print("ERROR: option_menu_set failed:", e)
        else:
            print("failed as expected", e)

    valid = MENU_TYPE(oi)

    #new_entry = 0x3333
    try:
        new_entry = ex.option_menu_get(oi['keyword'])
        if not valid:
            print("\nvbi_export_option_menu_get: unexpected success.")
    except Zvbi.ExportError as e:
        if valid:
            print("\nvbi_export_option_menu_get failed:", e)
        #if new_entry != 0x33333:
        #    raise Exception("vbi_export_option_menu_get failed, "+
        #                    "but modified destination to new_current\n")

    if ((oi['type'] == Zvbi.VBI_OPTION_BOOL) or
        (oi['type'] == Zvbi.VBI_OPTION_INT)):
        if MENU_TYPE(oi):
            if not (new_current == oi['menu'][new_entry]):
                raise Exception("FAILED")
        else:
                test_modified(oi, current, new_current)

        current = new_current
        print_current(oi, new_current)

    elif oi['type'] == Zvbi.VBI_OPTION_REAL:
        if MENU_TYPE(oi):
            # XXX unsafe
            if not (new_current == oi['menu'][new_entry]):
                raise Exception("FAILED")
        else:
            test_modified(oi, current, new_current)

        current = new_current
        print_current(oi, new_current)

    elif oi['type'] == Zvbi.VBI_OPTION_MENU:
        current = new_current
        print_current(oi, new_current)

    else:
        raise Exception("Unknown type:" + oi['type'])
    return current


def dump_option_info(ex, oi):
    type_str = TypeStr.get(oi['type'])
    if not type_str:
        raise Exception("invalid type in", str(oi))

    if not oi.get('label'):
        oi['label'] = "(null)";
    if not oi.get('tooltip'):
        oi['tooltip'] = "(null)";

    print("  * type=%s keyword=%s label=\"%s\" tooltip=\"%s\"" %
          (type_str, oi['keyword'], oi['label'], oi['tooltip']))

    keyword_check(oi['keyword'])

    if ((oi['type'] == Zvbi.VBI_OPTION_BOOL) or
        (oi['type'] == Zvbi.VBI_OPTION_INT)):
        BOUNDS_CHECK(oi)
        if MENU_TYPE(oi):
            print("    %d menu entries, default=%d: " %
                  (oi['max'] - oi['min'] + 1, oi['def']), end='')

            for i in range(0, oi['max'] + 1 - oi['min']):
                print("%d%s" % (oi['menu'][i],
                                (", " if (i < oi['max']) else "")), end='')
            print("")
        else:
            print("    default=%d, min=%d, max=%d, step=%d" %
                   (oi['def'], oi['min'], oi['max'], oi['step']))

        val = ex.option_get(oi['keyword'])

        print_current(oi, val)
        if opt.check:
            if MENU_TYPE(oi):
                val = test_set_entry(ex, oi, val, oi['min'])
                val = test_set_entry(ex, oi, val, oi['max'])
                val = test_set_entry(ex, oi, val, oi['min'] - 1)
                val = test_set_entry(ex, oi, val, oi['max'] + 1)
                val = test_set_int(ex, oi, val, oi['menu'][oi['min']])
                val = test_set_int(ex, oi, val, oi['menu'][oi['max']])
                val = test_set_int(ex, oi, val, oi['menu'][oi['min']] - 1)
                val = test_set_int(ex, oi, val, oi['menu'][oi['max']] + 1)
            else:
                val = test_set_entry(ex, oi, val, 0)
                val = test_set_int(ex, oi, val, oi['min'])
                val = test_set_int(ex, oi, val, oi['max'])
                val = test_set_int(ex, oi, val, oi['min'] - 1)
                val = test_set_int(ex, oi, val, oi['max'] + 1)

    elif oi['type'] == Zvbi.VBI_OPTION_REAL:
        BOUNDS_CHECK(oi)
        if MENU_TYPE(oi):
            print("    %d menu entries, default=%d: " %
                   (oi['max'] - oi['min'] + 1, oi['def']))
            for i in range(0, oi['max'] + 1 - oi['min']):
                print("%f%s" % (oi['menu'][i],
                                (", " if (i < oi['max']) else "")))
        else:
            print("    default=%f, min=%f, max=%f, step=%f" %
                   (oi['def'], oi['min'], oi['max'], oi['step']))

        val = ex.option_get(oi['keyword'])

        print_current(oi, val)
        if opt.check:
            if MENU_TYPE(oi):
                test_set_entry(ex, oi, val, oi['min'])
                test_set_entry(ex, oi, val, oi['max'])
                test_set_entry(ex, oi, val, oi['min'] - 1)
                test_set_entry(ex, oi, val, oi['max'] + 1)
                test_set_real(ex, oi, val, oi['menu'][oi['min']])
                test_set_real(ex, oi, val, oi['menu'][oi['max']])
                test_set_real(ex, oi, val, oi['menu'][oi['min']] - 1)
                test_set_real(ex, oi, val, oi['menu'][oi['max']] + 1)
            else:
                test_set_entry(ex, oi, val, 0)
                test_set_real(ex, oi, val, oi['min'])
                test_set_real(ex, oi, val, oi['max'])
                test_set_real(ex, oi, val, oi['min'] - 1)
                test_set_real(ex, oi, val, oi['max'] + 1)

    elif oi['type'] == Zvbi.VBI_OPTION_STRING:
        if MENU_TYPE(oi):
            STRING_CHECK(oi)
            print("    %d menu entries, default=%d: " %
                   (oi['max'] - oi['min'] + 1, oi['def']))
            for i in range(0, oi['max'] + 1 - oi['min']):
                print("%s%s" % (oi['menu'][i],
                                (", " if (i < oi['max']) else "")))
        else:
            print("    default=\"%s\"" % oi['def'])

        val = ex.option_get(oi['keyword'])

        print("    current value=\"%s\"" % val)
        if opt.check:
            print("    try to set \"foobar\": ", end='')
            try:
                ex.option_set(oi['keyword'], "foobar")
                printf("success.")
            except Zvbi.ExportError as e:
                print("failed:", e)

            val = ex.option_get(oi['keyword'])
            print("    current value=\"%s\"" % val)

    elif oi['type'] == Zvbi.VBI_OPTION_MENU:
            print("    %d menu entries, default=%d: " %
                   (oi['max'] - oi['min'] + 1, oi['def']), end='')
            for i in range(0, oi['max'] + 1 - oi['min']):
                print("%s%s" % (oi['menu'][i],
                                (", " if (i < oi['max']) else "")), end='')

            print("")
            val = ex.option_get(oi['keyword'])

            print_current(oi, val)
            if opt.check:
                val = test_set_entry(ex, oi, val, oi['min'])
                val = test_set_entry(ex, oi, val, oi['max'])
                val = test_set_entry(ex, oi, val, oi['min'] - 1)
                val = test_set_entry(ex, oi, val, oi['max'] + 1)

    else: # unreachable
        raise Exception("unknown type")

def list_options(ex):
    print("  List of options:")

    i = 0
    while True:
        try:
            oi = ex.option_info_enum(i)
            if not oi.get('keyword'):
                raise Exception("invalid option info:" + str(oi))

            foo = ex.option_info_keyword(oi['keyword'])
            dump_option_info(ex, oi)
            i += 1

        except StopIteration:
            break


def list_modules():
    print("List of export modules:")

    i = 0
    while True:
        try:
            xi = Zvbi.Export.info_enum(i)
            if not xi['keyword']:
                raise Exception()
            foo = Zvbi.Export.info_keyword(xi['keyword'])

            if not xi.get('label'):
                xi['label'] = "(null)"
            if not xi.get('tooltip'):
                xi['tooltip'] = "(null)"
            if not xi.get('mime_type'):
                xi['mime_type'] = "(null)"
            if not xi.get('extension'):
                xi['extension'] = "(null)"

            print(("* keyword=%s label=\"%s\"\n" +
                   "  tooltip=\"%s\" mime_type=%s extension=%s") %
                   (xi['keyword'], xi['label'],
                    xi['tooltip'], xi['mime_type'], xi['extension']))

            keyword_check(xi['keyword'])

            ex = Zvbi.Export(xi['keyword'])

            foo = ex.info_export()
            if not foo:
                raise Exception()

            list_options(ex)

            ex = None
            i += 1
        except StopIteration:
            break

        print("-- end of list --")


def ParseCmdOptions():
    global opt
    parser = argparse.ArgumentParser(description='List all export modules and options')
    parser.add_argument("--check", action='store_true', default=False, help="enable internal testing")
    opt = parser.parse_args()


ParseCmdOptions()
list_modules()



#!/usr/bin/python3
#
#  libzvbi test
#
#  Copyright (C) 2000-2006 Michael H. Schimek
#  Perl Port: Copyright (C) 2007 Tom Zoerner
#  Python Port: Copyright (C) 2020 Tom Zoerner
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# Description:
#
#   Example for the use of class Zvbi.Capture. The script captures VBI
#   data from a device and slices it. The result can be dumped for the
#   various data services in form of hex-string plus roughly decoded text
#   (where applicable) for inspection of incoming data. Altnernatively,
#   output can be written to STDOUT in binary format for further
#   processing (decoding) by piping the data into one of the following
#   example scripts. Call with option --help for a list of options.
#
#   (This is a translation of test/capture.c in the libzvbi package.)

import argparse
import sys
import re
import Zvbi

opt = None

#extern void
#vbi_capture_set_log_fp         (vbi_capture *          capture,
#                                FILE *                 fp)
#extern vbi_bool vbi_capture_force_read_mode

#
#  Dump
#

def PIL(day, mon, hour, minute):
    return ((day << 15) + (mon << 11) + (hour << 6) + (minute << 0))


def dump_pil(pil):
    day = pil >> 15
    mon = (pil >> 11) & 0xF
    hour = (pil >> 6) & 0x1F
    minute = pil & 0x3F

    if pil == PIL(0, 15, 31, 63):
        print(" PDC: Timer-control (no PDC)")
    elif pil == PIL(0, 15, 30, 63):
        print(" PDC: Recording inhibit/terminate")
    elif pil == PIL(0, 15, 29, 63):
        print(" PDC: Interruption")
    elif pil == PIL(0, 15, 28, 63):
        print(" PDC: Continue")
    elif pil == PIL(31, 15, 31, 63):
        print(" PDC: No time")
    else:
        print(" PDC: %05x, YYYY-%02d-%02d %02d:%02d" %
              (pil, mon, day, hour, minute))


pr_label = ""
tmp_label = ""

def decode_vps(buf):
    global pr_label
    global tmp_label

    if not opt.dump_vps:
        return

    print("\nVPS:")

    c = Zvbi.rev8(buf[1])

    if (c & 0x80):
        pr_label = tmp_label
        tmp_label = ""

    c &= 0x7F
    if (c >= 0x20) and (c < 0x7f):
        tmp_label += chr(c)
    else:
        tmp_label += '.'

    print(" 3-10: %02x %02x %02x %02x %02x %02x %02x %02x (\"%s\")" %
          (buf[0], buf[1], buf[2], buf[3], buf[4], buf[5], buf[6], buf[7], pr_label))

    pcs = buf[2] >> 6

    cni = (  ((buf[10] & 3) << 10)
           + ((buf[11] & 0xC0) << 2)
           + ((buf[8] & 0xC0) << 0)
           + (buf[11] & 0x3F))

    pil = ((buf[8] & 0x3F) << 14) + (buf[9] << 6) + (buf[10] >> 2)

    pty = buf[12]

    print(" CNI: %04x PCS: %d PTY: %d " % (cni, pcs, pty), end='')

    dump_pil(pil)


def decode_wss_625(buf):
    formats = (
            "Full format 4:3, 576 lines",
            "Letterbox 14:9 centre, 504 lines",
            "Letterbox 14:9 top, 504 lines",
            "Letterbox 16:9 centre, 430 lines",
            "Letterbox 16:9 top, 430 lines",
            "Letterbox > 16:9 centre",
            "Full format 14:9 centre, 576 lines",
            "Anamorphic 16:9, 576 lines"
    )
    subtitles = (
            "none",
            "in active image area",
            "out of active image area",
            "<invalid>"
    )

    if opt.dump_wss:
        g1 = buf[0] & 15
        parity = g1
        parity ^= parity >> 2
        parity ^= parity >> 1
        g1 &= 7

        print("WSS PAL: ", end='')
        if not (parity & 1):
            print("<parity error> ", end='')

        print(("%s; %s mode; %s colour coding; %s helper; "+
              "reserved b7=%d; %s Teletext subtitles; "+
              "open subtitles: %s; %s surround sound; "+
              "copyright %s; copying %s") %
              ( formats[g1],
               ("film" if (buf[0] & 0x10) else "camera"),
               ("MA/CP" if (buf[0] & 0x20) else "standard"),
               ("modulated" if (buf[0] & 0x40) else "no"),
               (buf[0] & 0x80) != 0,
               ("have" if (buf[1] & 0x01) else "no"),
               subtitles[(buf[1] >> 1) & 3],
               ("have" if (buf[1] & 0x08) else "no"),
               ("asserted" if (buf[1] & 0x10) else "unknown"),
               ("restricted" if (buf[1] & 0x20) else "not restricted")
             ))


def decode_wss_cpr1204(buf):
    if opt.dump_wss:
        poly = (1 << 6) + (1 << 1) + 1
        g = (buf[0] << 12) + (buf[1] << 4) + buf[2]
        crc = g | (((1 << 6) - 1) << (14 + 6))

        for j in range(14 + 6 - 1, -1, -1):
            if (crc & ((1 << 6) << j)):
                crc ^= poly << j

        print("WSS CPR >> g=%08x crc=%08x" % (g, crc), file=sys.stderr)


def decode_sliced(sliced_buf):
    if opt.dump_sliced:
        print("Sliced time: %f" % sliced_buf.timestamp)

        for data, slc_id, line in sliced_buf:
            print("%04x %3d > " % (slc_id, line), end='')

            for d in data:
                print("%02x " % d, end='')

            print(" ", end='')

            astr = Zvbi.unpar_str(data, '.')
            astr = re.sub(r'[\x00-\x1F\x7F]', '.', astr.decode('ISO-8859-1'))
            print(astr)

    for data, slc_id, line in sliced_buf:
        if slc_id == 0:
            pass # nop
        elif not (slc_id & Zvbi.VBI_SLICED_VPS) == 0:
          decode_vps(data)
        elif not (slc_id & Zvbi.VBI_SLICED_TELETEXT_B) == 0:
          pass # Use ./decode instead.
        elif not (slc_id & Zvbi.VBI_SLICED_CAPTION_525) == 0:
          pass # Use ./decode instead.
        elif not (slc_id & Zvbi.VBI_SLICED_CAPTION_625) == 0:
          pass # Use ./decode instead.
        elif not (slc_id & Zvbi.VBI_SLICED_WSS_625) == 0:
          decode_wss_625(data)
        elif not (slc_id & Zvbi.VBI_SLICED_WSS_CPR1204) == 0:
            decode_wss_cpr1204(data)
        else:
            print("Oops. Unhandled vbi service %08x\n" % slc_id, file=sys.stderr)

#
#  Sliced, binary
#

# hysterical compatibility
ServiceWidth = {
        Zvbi.VBI_SLICED_TELETEXT_B:  (42, 0),
        Zvbi.VBI_SLICED_CAPTION_625: (2, 1),
        Zvbi.VBI_SLICED_VPS:         (13, 2),
        Zvbi.VBI_SLICED_WSS_625:     (2, 3),
        Zvbi.VBI_SLICED_WSS_CPR1204: (3, 4),
        Zvbi.VBI_SLICED_CAPTION_525: (2, 7),
}

last_ts = 0.0

def binary_sliced(sliced_buf):
    global last_ts

    ts = sliced_buf.timestamp - last_ts if (last_ts > 0.0) else 0.04
    outfile.write(bytes("%f\n" % ts, 'ascii'))

    outfile.write(bytes([len(sliced_buf)]))

    for data, slc_id, line in sliced_buf:
        if ServiceWidth.get(slc_id) and (ServiceWidth.get(slc_id)[0] > 0):
            outfile.write(bytes([ServiceWidth.get(slc_id)[1],
                                 line & 0xFF,
                                 line >> 8]))
            data_len = ServiceWidth.get(slc_id)[0]
            outfile.write(data[0 : data_len])
            last_ts = sliced_buf.timestamp

    outfile.flush()


def binary_ts_pes(packet, user_data=None):
    outfile.write(packet)
    outfile.flush()
    return 1


def mainloop(cap):
    dump = (opt.dump_wss or opt.dump_vps or opt.dump_sliced)
    err_cnt = 0

    while True:
        try:
            if opt.read:
                raw_buf, sliced_buf = cap.read(1000)
            else:
                raw_buf, sliced_buf = cap.pull(1000)
            err_cnt = 0
        except Zvbi.CaptureError as e:
            if not opt.ignore_error:
                print("Capture error:", e, file=sys.stderr)
                err_cnt += 1  # ignore occasional singular errors
                if err_cnt >= 2:
                    break
                continue
        except Zvbi.CaptureTimeout:
            if not opt.ignore_error:
                print("Capture timeout", file=sys.stderr)
            continue

        if False:
            print(".", file=sys.stderr)
            sys.stderr.flush()

        if dump:
            decode_sliced(sliced_buf)

        if opt.sliced:
            binary_sliced(sliced_buf)

        if opt.pes or opt.ts:
            # XXX shouldn't use system time
            pts = int(sliced_buf.timestamp * 90000.0)
            services = (Zvbi.VBI_SLICED_TELETEXT_B |
                        Zvbi.VBI_SLICED_VPS |
                        Zvbi.VBI_SLICED_CAPTION_625 |
                        Zvbi.VBI_SLICED_WSS_625)
            mx.feed(services, sliced_buf=sliced_buf, pts=pts)


#static const char short_options[] = "123cd:elnpr:stvPT"

def ParseCmdOptions():
    global opt
    parser = argparse.ArgumentParser(description='ZVBI capturing example')
    parser.add_argument("--device", type=str, default="/dev/dvb/adapter0/demux0", help="Path to video capture device")  # dev_name,
    parser.add_argument("--pid", type=int, default=0, help="Teletext channel PID for DVB")
    parser.add_argument("--ignore-error", action='store_true', default=False, help="Silently ignore device errors and timeout")
    parser.add_argument("--dump-ttx", action='store_true', default=False, help="Capture and dump teletext packets")
    parser.add_argument("--dump-xds", action='store_true', default=False, help="Capture and dump CC XDS packets")
    parser.add_argument("--dump-cc", action='store_true', default=False, help="Capture and dump CC packets")
    parser.add_argument("--dump-wss", action='store_true', default=False, help="Capture and dump WSS")
    parser.add_argument("--dump-vps", action='store_true', default=False, help="Capture and dump VPS data")
    parser.add_argument("--dump-sliced", action='store_true', default=False, help="Capture and all VBI services")
    parser.add_argument("--pes", action='store_true', default=False, help="Write output as PES DVB stream")   # bin_pes,
    parser.add_argument("--ts", action='store_true', default=False, help="Write output as TS DVB stream")   # bin_ts,
    parser.add_argument("--sliced", action='store_true', default=False, help="Write binary output, for piping into decode.py")   # bin_sliced,
    parser.add_argument("--read", action='store_true', default=False, help="Use \"read\" capture method instead of \"pull\"")   # do_read,
    parser.add_argument("--strict", type=int, default=0, help="Use strict mode 0,1,2 for adding VBI services")
    #parser.add_argument("--desync", action='store_true', default=False)
    #parser.add_argument("--sim", action='store_true', default=False)   # do_sim,
    parser.add_argument("--pal", action='store_true', default=False, help="Assume PAL video norm (bktr driver only)")   # scanning_pal,
    parser.add_argument("--ntsc", action='store_true', default=False, help="Assume NTSC video norm (bktr driver only)")   # scanning_ntsc,
    #parser.add_argument("--v4l", action='store_true', default=False)   # api_v4l,
    parser.add_argument("--v4l2", action='store_true', default=False, help="Using analog driver interface")
    #parser.add_argument("--v4l2-read", action='store_true', default=False)   # api_v4l2, # FIXME
    #parser.add_argument("--v4l2-mmap", action='store_true', default=False)   # api_v4l2, # FIXME
    parser.add_argument("--verbose", action='store_true', default=False, help="Enable trace output in the library")
    opt = parser.parse_args()

    if opt.v4l2 and (opt.pid != 0):
        print("Options --v4l2 and --pid are mutually exclusive", file=sys.stderr)
        sys.exit(1)
    if not opt.v4l2 and (opt.pid == 0) and ("dvb" in opt.device):
        print("WARNING: DVB devices require --pid parameter", file=sys.stderr)


def main_func():
    if opt.pal:
        scanning = 625
    elif opt.ntsc:
        scanning = 525
    else:
        scanning = 0

    dump = (opt.dump_wss or opt.dump_vps or opt.dump_sliced)

    if opt.dump_ttx or opt.dump_cc or opt.dump_xds:
        print("Teletext, CC and XDS decoding are no longer supported by this tool.\n" +
              "Run  ./capture --sliced | ./decode --ttx --cc --xds  instead.\n", file=sys.stderr)
        exit(-1)

    if opt.sliced:
        if opt.pes or opt.ts or dump:
            print("WARNING: combining --sliced with --pes, --ts or --dump* will garble output", file=sys.stderr)
    elif opt.pes or opt.ts:
        if opt.sliced or dump:
            print("WARNING: combining --pes/ts with --sliced or --dump* will garble output", file=sys.stderr)

    services = (Zvbi.VBI_SLICED_VBI_525 |
                Zvbi.VBI_SLICED_VBI_625 |
                Zvbi.VBI_SLICED_TELETEXT_B |
                Zvbi.VBI_SLICED_CAPTION_525 |
                Zvbi.VBI_SLICED_CAPTION_625 |
                Zvbi.VBI_SLICED_VPS |
                Zvbi.VBI_SLICED_WSS_625 |
                Zvbi.VBI_SLICED_WSS_CPR1204)

    if False: #opt.sim:
        #TODO cap = Zvbi.sim_new (scanning, services, 0, !opt.desync)
        #par = cap.parameters()
        pass
    elif opt.v4l2 or (opt.pid == 0 and not "dvb" in opt.device):
        cap = Zvbi.Capture.Analog(opt.device, services=services, scanning=scanning,
                                  strict=opt.strict, trace=opt.verbose, buffers=5)
        par = cap.parameters()
    else:
        cap = Zvbi.Capture.Dvb(opt.device, dvb_pid=opt.pid, trace=opt.verbose)
        par = cap.parameters()

    if opt.verbose > 1:
        Zvbi.set_log_on_stderr(Zvbi.VBI_LOG_ERROR |
                               Zvbi.VBI_LOG_WARNING |
                               Zvbi.VBI_LOG_INFO)

    if opt.pid == -1:
        if par.sampling_format != Zvbi.VBI_PIXFMT_YUV420:
            print("Unexpected sampling format:%d in capture parameters"
                    % par.sampling_format, file=sys.stderr)
            exit(-1)

    global mx
    if opt.pes:
        mx = Zvbi.DvbMux(pes=True, callback=binary_ts_pes)
        mx.set_pes_packet_size (0, 8* 184)
    elif opt.ts:
        mx = Zvbi.DvbMux(ts_pid=999, callback=binary_ts_pes)
        mx.set_pes_packet_size (0, 8* 184)

    global outfile
    outfile = open(sys.stdout.fileno(), "wb")

    mainloop(cap)

# main
try:
    ParseCmdOptions()
    main_func()
except (KeyboardInterrupt, BrokenPipeError):
    pass



# ----------------------------------------------------------------------------
# Copyright (C) 2007-2009,-2020 T. Zoerner
# ----------------------------------------------------------------------------
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation (see http://www.opensource.org/licenses/)
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# ----------------------------------------------------------------------------

from setuptools import setup, Extension
from distutils.command.install import install

import os
import os.path
import sys
import platform
import shutil
import re

if sys.version_info[0] != 3:
    print("This script requires Python 3")
    exit()

# ----------------------------------------------------------------------------

class MyClean(install):
    cwd = os.path.abspath(os.path.dirname(__file__))
    def rmfile(self, apath):
        p = os.path.join(MyClean.cwd, apath)
        if os.path.isfile(p):
            os.remove(p)
    def rmtree(self, apath):
        p = os.path.join(MyClean.cwd, apath)
        if os.path.isdir(p):
            shutil.rmtree(p)
    def run(self):
        # files created by build stage
        self.rmtree('build')
        # files created by test stage
        self.rmtree('Zvbi.egg-info')
        self.rmtree('__pycache__')
        for name in os.listdir(MyClean.cwd):
            if re.match(r"^.*\.so$", name):
                os.remove(os.path.join(MyClean.cwd, name))
        self.rmfile('core')
        # files created by sdist stage
        self.rmtree('dist')

# ----------------------------------------------------------------------------

this_directory = os.path.abspath(os.path.dirname(__file__))

# platform-specific build parameters
extrasrc = []       # extra source modules to compile
extrainc = []       # include path
extradef = []       # extra copile-switches
extralibdirs = []   # extra library search path
extralibs = []      # extra libraries to link against

# BSD requires listing libraries that libzvbi depends on
if re.match(r'bsd$', platform.system(), flags=re.IGNORECASE):
    extralibs += ['pthread', 'png', 'z']

# ----------------------------------------------------------------------------

# Assemble main README from doc/README.rsr plus all but the header of API doc
with open(os.path.join(this_directory, 'doc/README.rst'), encoding='utf-8') as fh:
    long_description = fh.read()
with open(os.path.join(this_directory, 'doc/ZVBI.rst'), encoding='utf-8') as fh:
    api_doc = fh.read()
    match = re.match(r"[\x00-\xff]*?(?=^SYNOPSIS$)", api_doc, re.MULTILINE)
    if match:
        long_description += "\n\n" + api_doc[match.end():]
    else:
        print("ERROR: Failed to find SYNOPSIS in ZVBI.rst", file=sys.stderr)

# Enable work-around for bugs in PyStructSequence_NewType (i.e. for
# creating named tuples in C module; causing crash in GC:
# "Fatal Python error: type_traverse() called for non-heap type")
# Known issue: https://bugs.python.org/issue28709 - fixed in 3.8
if (sys.version_info[0] == 3) and (sys.version_info[1] < 8):
    extradef += [('NAMED_TUPLE_GC_BUG', 1)]

# Definition of the "Zvbi" extension module
ext = Extension('Zvbi',
                sources       = ['src/zvbi.c',
                                 'src/zvbi_proxy.c',
                                 'src/zvbi_capture.c',
                                 'src/zvbi_capture_buf.c',
                                 'src/zvbi_raw_dec.c',
                                 'src/zvbi_raw_params.c',
                                 'src/zvbi_service_dec.c',
                                 'src/zvbi_page.c',
                                 'src/zvbi_event_types.c',
                                 'src/zvbi_callbacks.c',
                                 'src/zvbi_export.c',
                                 'src/zvbi_search.c',
                                 'src/zvbi_dvb_mux.c',
                                 'src/zvbi_dvb_demux.c',
                                 'src/zvbi_idl_demux.c',
                                 'src/zvbi_pfc_demux.c',
                                 'src/zvbi_xds_demux.c',
                                ] + extrasrc,
                include_dirs  = ['src'] + extrainc,
                define_macros = extradef,
                libraries     = ['zvbi'] + extralibs,
                library_dirs  = extralibdirs,
                #undef_macros  = ["NDEBUG"]   # for debug build only
               )

# Package definition
setup(name='Zvbi',
      version='0.1.1',
      description='Interface to the Zapping VBI decoder library',
      long_description=long_description,
      long_description_content_type="text/x-rst",
      author='T. Zoerner',
      author_email='tomzo@users.sourceforge.net',
      url='https://github.com/tomzox/Python-ZVBI',
      license = "GNU GPLv2",
      classifiers=[
          'Development Status :: 4 - Beta',
          "Programming Language :: C",
          "Programming Language :: Python :: 3",
          'Topic :: Multimedia :: Video :: Capture',
          'Intended Audience :: Developers',
          "Operating System :: POSIX :: Linux",
          "Operating System :: POSIX :: BSD",
          "License :: OSI Approved :: GNU General Public License v2 (GPLv2)"
         ],
      keywords="teletext, videotext, closed-caption, VPS, WSS, VBI, DVB, video, capture, decoder, libzvbi, ETSI",
      platforms=['posix'],
      ext_modules=[ext],
      cmdclass={'clean': MyClean},
      python_requires='>=3.2',
     )



#!/usr/bin/python3
#
#  libzvbi test of page export in different formats
#
#  Copyright (C) 2000, 2001 Michael H. Schimek
#  Perl Port: Copyright (C) 2006, 2007 Tom Zoerner
#  Python Port: Copyright (C) 2020 Tom Zoerner
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# Description:
#
#   Example for the use of export actions in class Zvbi.Export. The script
#   captures from a device until the page specified on the command line is
#   found and then exports the page content in a requested format.
#   Alternatively, the script can be used to continuously export a single
#   or all received pages. Examples:
#
#     ./export.py text 100       # teletext page 100 as text
#     ./export.py text all       # continuously all teletext pages
#     ./export.py --loop text 1  # continuously closed caption
#     ./export.py "png;reveal=1" 100 > page_100.png
#
#   Use ./explist.py for listing supported export formats (aka "modules")
#   and possible options. Note options are appended to the module name,
#   separated by semicolon as shown in the second example.
#
#   (This is a direct translation of test/export.c in libzvbi.)

import sys
import re
import argparse
import Zvbi

opt = None
vtdec = None
ex = None
extension = "dat"
quit = False
cr = '\r' if sys.stderr.isatty() else "\n"


def ttx_handler(pg_type, ev, user_data=None):
    global quit

    print("%cPage %03x.%02x " % (cr, ev.pgno, ev.subno & 0xFF), end='', file=sys.stderr)

    if ev.pgno != opt.page and opt.page != -1:
        return

    if sys.stderr.isatty():
        print("", file=sys.stderr)
    print("Saving page %03X..." % ev.pgno, end=('' if opt.to_file else '\n'), file=sys.stderr)

    # Fetching & exporting here is a bad idea,
    # but this is only a test.
    with vtdec.fetch_vt_page(ev.pgno, ev.subno) as page:
        try:
            if opt.to_file:
                name = ("ttx-%03x-%02x.%s" % (ev.pgno, ev.subno, extension))
                ex.to_file(page, name)
            else:
                sys.stdout.flush()
                fd = sys.stdout.fileno()
                ex.to_stdio(page, fd)

        except Zvbi.ExportError as e:
            print("export failed:", e, file=sys.stderr)
            sys.exit(-1)

        # for test purpose only
        str = ex.to_memory(page)
        #print(str)
        str = None

        if not opt.loop and opt.page != -1:
            print("done", file=sys.stderr)
            quit = True


def cc_handler(pg_type, ev, user_data=None):
    global quit

    print("%cCC page %d " % (cr, ev.pgno), end='', file=sys.stderr)
    if not ev.pgno == opt.page:
        return

    with vtdec.fetch_cc_page(ev.pgno) as page:
        try:
            if opt.to_file:
                name = ("cc-%d.%s" % (ev.pgno, extension))
                ex.to_file(page, name)
            else:
                sys.stdout.flush()
                fd = sys.stdout.fileno()
                ex.to_stdio(page, fd)

        except Zvbi.ExportError as e:
            print("export failed:", e, file=sys.stderr)
            sys.exit(-1)

    if not opt.loop and opt.page != -1:
        print("done", file=sys.stderr)
        quit = True


def pes_mainloop():
    infile = open(sys.stdin.fileno(), "rb")
    dvb = Zvbi.DvbDemux()

    while not quit:
        buf = infile.read(2048)
        if len(buf) == 0:
            break

        dvb.feed(buf)
        for sliced_buf in dvb:
            vtdec.decode(sliced_buf)

    if not quit:
        print("\rEnd of stream, page %03x not found" % opt.page, file=sys.stderr)


def dev_mainloop():
    global cap

    if opt.v4l2 or (opt.pid == 0 and not "dvb" in opt.device):
        opt_services = (Zvbi.VBI_SLICED_TELETEXT_B |
                        Zvbi.VBI_SLICED_CAPTION_525 |
                        Zvbi.VBI_SLICED_CAPTION_625);
        opt_buf_count = 5
        opt_strict = 0

        cap = Zvbi.Capture.Analog(opt.device, services=opt_services,
                                  buffers=opt_buf_count, strict=opt_strict,
                                  trace=opt.verbose)
    else:
        cap = Zvbi.Capture.Dvb(opt.device, dvb_pid=opt.pid, trace=opt.verbose)

    while not quit:
        try:
            sliced_buf = cap.pull_sliced(1000)
            vtdec.decode(sliced_buf)
        except Zvbi.CaptureError as e:
            print("Capture error:", e)
        except Zvbi.CaptureTimeout:
            print("Capture timeout:")

    if not quit:
        print("\rEnd of stream, page %03x not found\n" % opt.page, file=sys.stderr)


def main_func():
    global extension
    global ex
    global vtdec

    try:
        ex = Zvbi.Export(opt.module)
    except Zvbi.ExportError as e:
        print("FATAL:", e, file=sys.stderr)
        exit(1)

    xi = ex.info_export()
    extension = xi['extension']
    re.sub(r',.*', '', extension)
    xi = None

    vtdec = Zvbi.ServiceDec()
    vtdec.event_handler_register(Zvbi.VBI_EVENT_TTX_PAGE, ttx_handler)
    vtdec.event_handler_register(Zvbi.VBI_EVENT_CAPTION, cc_handler)

    if opt.pes:
        pes_mainloop()
    else:
        dev_mainloop()


def ttx_pgno(val_str):
    if val_str == "all":
        val = -1
    else:
        val = int(val_str, 16)
        if not ((val >= 0x100 and val <= 0x8FF) or (val >= 1 and val <= 8)):
            raise ValueError("Teletext page not in range 100-8FF")
    return val


def ParseCmdOptions():
    global opt

    if len(sys.argv) <= 1:
        print("Usage: %s \"module[;option=value]\" pgno >file" % sys.argv[0],
              "Where: module is eg. \"text\" or \"ppm\"",
              "       pgno is eg. 100 (hex) or \"all\"",
              "Run explist.py for a full list of modules and options",
              "Use --help to get list of command line options",
              sep='\n', file=sys.stderr)
        sys.exit(1)

    parser = argparse.ArgumentParser(description='Search teletext for an entered string')
    parser.add_argument("--device", type=str, default="/dev/dvb/adapter0/demux0", help="Path to video capture device")
    parser.add_argument("--pid", type=int, default=0, help="Teletext channel PID for DVB")
    parser.add_argument("--v4l2", action='store_true', default=False, help="Using analog driver interface")
    parser.add_argument("--pes", action='store_true', default=False, help="Read DVB PES input stream from STDIN")
    parser.add_argument("--loop", action='store_true', default=False, help="Do not exit after exporting; repeat for every reception")
    parser.add_argument("--to-file", action='store_true', default=False, help="Store to file named ttx-PGNO-SUBPG.dat or CC-PGNO.dat")
    parser.add_argument("--verbose", action='store_true', default=False, help="Enable trace output in the library")
    parser.add_argument("module", type=str, action='store', help="Export format, optionally followed by option assignments: FORMAT;OPTION=VALUE")
    parser.add_argument("page", type=ttx_pgno, action='store', help="Page number to export")
    opt = parser.parse_args()

    if not opt.pes:
        if opt.v4l2 and (opt.pid != 0):
            print("Options --v4l2 and --pid are mutually exclusive", file=sys.stderr)
            sys.exit(1)
        if not opt.v4l2 and (opt.pid == 0) and ("dvb" in opt.device):
            print("WARNING: DVB devices require --pid parameter", file=sys.stderr)
    else:
        if opt.v4l2 or (opt.pid != 0):
            print("Options -pes, --v4l2 and --pid are mutually exclusive", file=sys.stderr)
            sys.exit(1)
        if sys.stdin.isatty():
            print("No PES data on stdin", file=sys.stderr)
            sys.exit(-1)

try:
    ParseCmdOptions()
    main_func()
except (KeyboardInterrupt, BrokenPipeError):
    pass



#!/usr/bin/python3
#
#  Minimal level 2.5 teletext browser
#
#  Copyright (C) 2007,2020 Tom Zoerner
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# Description:
#
#   This script is an example for the use of classes Zvbi.Page and Zvbi.Export
#   for rendering teletext pages. The script captures teletext from a given
#   device and renders selected teletext pages in a simple GUI using TkInter.

import sys
import math
import argparse
import tkinter.font as tkf
from tkinter import *
import Zvbi

#
# Global variables
#
pxc = None
cap = None
vtdec = None
pg_disp = -1
pg_sched = 0x100

tk = None
canvas = None
img_xpm = None
redraw = False

#
# This callback is invoked by the teletext decoder for every ompleted page.
# The function updates the page number display on top of the window and
# updates the display if the scheduled page has been captured.
#
def pg_handler(pgtype, ev, user_data=None):
    pg_lab.set("Page %03x.%02x " % (ev.pgno, ev.subno & 0xFF))

    global redraw
    if ev.pgno == pg_sched:
        redraw = True

#
# This function is called every 10ms to capture VBI data.
# VBI frames are sliced and forwarded to the teletext decoder.
#
def cap_frame():
    global tid_polling
    global redraw

    try:
        sliced_buf = cap.pull_sliced(10)
        vtdec.decode(sliced_buf)
    except Zvbi.CaptureError as e:
        print("Capture error:", e, file=sys.stderr)
    except Zvbi.CaptureTimeout:
        pass

    if redraw:
        pg_display()
        redraw = False

    # reschedule the handler
    tid_polling = tk.after(10, cap_frame)

#
# This function is called once during start-up to initialize the
# device capture context and the teletext decoder
#
def cap_init():
    global cap
    global vtdec
    global tid_polling

    opt.verbose = False
    if opt.v4l2 or (opt.pid == 0 and not "dvb" in opt.device):
        opt_buf_count = 5
        opt_services = Zvbi.VBI_SLICED_TELETEXT_B
        opt_scanning = (525 if opt.ntsc else (625 if opt.pal else 0))
        opt_strict = 0
        cap = Zvbi.Capture.Analog(opt.device, services=opt_services, scanning=opt_scanning,
                                  buffers=opt_buf_count, strict=opt_strict, trace=opt.verbose)
    else:
        cap = Zvbi.Capture.Dvb(opt.device, dvb_pid=opt.pid, trace=opt.verbose)

    vtdec = Zvbi.ServiceDec()
    vtdec.event_handler_register(Zvbi.VBI_EVENT_TTX_PAGE, pg_handler)

    # install a Tk event handler for capturing in the background
    #tk.fileevent(cap.get_fd(), 'readable', cap_frame) # not supported in Python [lame]
    tid_polling = tk.after_idle(cap_frame)

#
# Render the teletext page using PPM (image) export
#
def pg_display_xpm():
    try:
        pg = vtdec.fetch_vt_page(pg_sched)
        (h, w) = pg.get_page_size()

        if False:
            # Pixmap class only exists in Perl -> XPM not supported in TkInter
            # suppress all XPM extensions because Pixmap can't handle them
            #ex = Zvbi.Export('xpm')
            #ex.option_set('creator', "")
            #ex.option_set('titled', 0)
            #ex.option_set('transparency', 0)
            #img_obj = Pixmap(data=img_data)
            pass

        elif False:
            # export page in PPM format
            ex = Zvbi.Export('ppm')
            img_data = ex.to_memory(pg)
            img_obj = tk.call("image", "create", "photo", "-data", img_data)

        else:
            # draw page to RGB canvas & convert to PPM
            # conversion of 8-bit palette image format into XPM is faster
            #fmt = Zvbi.VBI_PIXFMT_PAL8
            fmt = Zvbi.VBI_PIXFMT_RGBA32_LE
            img_canvas = pg.draw_vt_page(fmt=fmt)
            img_data = pg.canvas_to_ppm(img_canvas, fmt, aspect=1)
            img_obj = tk.call("image", "create", "photo", "-data", img_data)
            # FIXME PhotoImage() generates blank (black) image
            #img_obj = PhotoImage(data=img_data, name="ttx")

        canvas.delete('all')
        canvas.create_image(0, 0, anchor=NW, image=img_obj)
        #canvas.configure(width=img_obj.width(), height=img_obj.height())
        w = tk.call("image", "width", img_obj)
        h = tk.call("image", "height", img_obj)
        canvas.configure(width=w, height=h)

        global pg_disp
        pg_disp = pg_sched

    except Zvbi.ServiceDecError:
        pass


#
# Return color name for the given 24-bit R/G/B value
#
def vbi_rgba(v):
    return ("#%02X%02X%02X" % (v&0xff, (v>>8)&0xff, (v>>16)&0xff))


#
# Render the teletext page using page text/property query
#
def pg_display_text():
    try:
        pg = vtdec.fetch_vt_page(pg_sched)

        (rows, columns) = pg.get_page_size()
        pal = pg.get_page_color_map()
        text = pg.get_page_text(' ')
        prop = pg.get_page_text_properties()

        fh = tk.call("font", "metrics", font, "-linespace")
        fw = font.measure("0")

        canvas.delete('all')
        i = 0
        for row in range(0, rows):
            for col in range(0, columns):
                pp = prop[i]
                canvas.create_rectangle(col * fw, row * fh,
                                        (col+1) * fw, (row+1) * fh,
                                        outline="",
                                        fill=vbi_rgba(pal[(pp>>8) & 0xFF]))
                canvas.create_text(col * fw, row * fh,
                                   text=text[i], anchor=NW, font=font,
                                   fill=vbi_rgba(pal[pp & 0xFF]))
                i += 1

        canvas.configure(width=(columns * fw), height=(rows * fh))
        global pg_disp
        pg_disp = pg_sched

    except (Zvbi.ServiceDecError, Zvbi.PageError):
        pass


#
# This function is invoked out of the capture event handler when the page
# which is scheduled for display has been captured.
#
def pg_display():
    if mode_xpm.get():
        pg_display_xpm()
    else:
        pg_display_text()

#
# This callback is invoked when the user clicks into the teletext page.
# If there's a page number of FLOF link under the mouse pointer, the
# respective page is scheduled for display.
#
def pg_link(x, y):
    global redraw
    global pg_sched

    if not pg_disp == -1:
        try:
            pg = vtdec.fetch_vt_page(pg_disp)
            if mode_xpm.get():
                # note: char width 12, char height 10*2 due to scaling in XPM conversion
                fh = 20
                fw = 12
            else:
                fh = tk.call("font", "metrics", font, "-linespace")
                fw = font.measure("0")

            link = pg.resolve_link(x // fw, y // fh)
            if link.type == Zvbi.VBI_LINK_PAGE:
                pg_sched = link.pgno
                dec_entry.set("%03X" % pg_sched)
            redraw = True
        except (Zvbi.ServiceDecError, Zvbi.PageError):
            pass

#
# This callback is invoked when the user hits the left/right buttons
# (actually this is redundant to the +/- buttons in the spinbox)
#
def pg_plus_minus(off):
    if off >= 0:
        off = 1
    else:
        off = 0xF9999999

    global pg_sched
    pg_sched = Zvbi.add_bcd(pg_sched, off)
    if pg_sched < 0x100:
        pg_sched = 0x899
    dec_entry.set("%03X" % pg_sched)

    global redraw
    redraw = True

#
# This callback is invoked when the user edits the page number
#
def pg_change():
    global redraw
    global pg_sched
    try:
        v = int(dec_entry.get())
        pg_sched = Zvbi.dec2bcd(v)
        redraw = True
    except ValueError:
        pass

#
# This callback is invoked when the user hits the "TOP" button
# to display the TOP page table
#
def pg_top_index():
    global pg_sched
    global redraw

    pg_sched = 0x900
    dec_entry.set("900")
    redraw = True

#
# This function is called once during start-up to create the GUI.
#
def gui_init():
    global tk
    global canvas
    global dec_entry
    global mode_xpm
    global pg_lab
    global font
    global redraw

    tk = Tk(className="browse-ttx")
    tk.wm_title('Teletext Level 2.5 Demo')
    tk.wm_resizable(0, 0)

    dec_entry = StringVar(tk, '100')
    mode_xpm = BooleanVar(tk, True)
    pg_lab = StringVar(tk, "Page 000.00")
    font = tkf.Font(family='courier', size=12)

    # frame holding control widgets at the top of the window
    wid_f1 = Frame(tk)
    wid_f1_sp = Spinbox(wid_f1, from_=100, to=899, width=5,
                        textvariable=dec_entry, command=pg_change)
    wid_f1_sp.bind('<Return>', lambda e: pg_change())
    wid_f1_sp.pack(side=LEFT, anchor=W)
    wid_f1_lab = Label(wid_f1, textvariable=pg_lab)
    wid_f1_lab.pack(side=LEFT)
    wid_f1.pack(side=TOP, fill=X)
    wid_f1_but1 = Button(wid_f1, text="<<", command=lambda: pg_plus_minus(-1), padx=1)
    wid_f1_but2 = Button(wid_f1, text=">>", command=lambda: pg_plus_minus(1), padx=1)
    wid_f1_but3 = Button(wid_f1, text="TOP", command=lambda: pg_top_index(), padx=1)
    wid_f1_but1.pack(side=LEFT, anchor=E)
    wid_f1_but2.pack(side=LEFT, anchor=E)
    wid_f1_but3.pack(side=LEFT, anchor=E)
    wid_f1_mode = Checkbutton(wid_f1, variable=mode_xpm, text="XPM",
                              command=lambda:pg_change())
    wid_f1_mode.pack(side=LEFT, anchor=E)

    # canvas for displaying the teletext page as image
    canvas = Canvas(tk, borderwidth=0, relief='flat', background='#000000')
    canvas.bindtags([canvas, 'all'])  # remove widget default bindings
    canvas.bind('<Key-q>', lambda e: sys.exit(0))
    canvas.bind('<Button-1>', lambda e: pg_link(e.x, e.y))
    canvas.pack(fill=BOTH)
    canvas.focus_set()

    redraw = False


def main_func():
    # create & display GUI
    gui_init()

    # start capturing teletext
    cap_init()

    # everything from here on is event driven
    tk.mainloop()


def ParseCmdOptions():
    global opt
    parser = argparse.ArgumentParser(description="Plotter of captured raw VBI data")
    parser.add_argument("--device", type=str, default="/dev/dvb/adapter0/demux0", help="Path to video capture device")
    parser.add_argument("--pid", type=int, default=0, help="Teletext channel PID for DVB")
    parser.add_argument("--v4l2", action='store_true', default=False, help="Using analog driver interface")
    parser.add_argument("--pal", action='store_true', default=False, help="Assume PAL video norm (bktr driver only)")
    parser.add_argument("--ntsc", action='store_true', default=False, help="Assume NTSC video norm (bktr driver only)")
    parser.add_argument("--verbose", action='store_true', default=False, help="Enable trace output in the library")
    opt = parser.parse_args()

    if opt.v4l2 and (opt.pid != 0):
        print("Options --v4l2 and --pid are mutually exclusive", file=sys.stderr)
        sys.exit(1)
    if not opt.v4l2 and (opt.pid == 0) and ("dvb" in opt.device):
        print("WARNING: DVB devices require --pid parameter", file=sys.stderr)


try:
    ParseCmdOptions()
    main_func()
except KeyboardInterrupt:
    pass



#!/usr/bin/python3
#
#  VBI proxy test client
#
#  Copyright (C) 2003,2004,2006,2007,2020 Tom Zoerner
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License version 2 as
#  published by the Free Software Foundation.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#
#  Description:
#
#   Example for the use of class Zvbi.Proxy. The script can capture either
#   from a proxy daemon or a local device. Continuously shows summary of
#   captured data on the terminal. Also allows changing services and
#   channels during capturing (e.g.  by entering "+ttx" or "-ttx" via
#   terminal at STDIN.) Run the script with option -help for a list of
#   supported command line options.
#
#   (This is a direct translation of test/proxy-test.c in libzvbi.)

import sys
import argparse
import struct
import re
import select
import Zvbi

# constants from linux/videodev2.h
VIDIOC_ENUMINPUT = 0xC050561A
VIDIOC_G_INPUT = 0x80045626
VIDIOC_S_INPUT = 0xC0045627
VIDIOC_G_FREQUENCY = 0xC02C5638
VIDIOC_S_FREQUENCY = 0x402C5639

all_services_625 =   ( Zvbi.VBI_SLICED_TELETEXT_B |
                       Zvbi.VBI_SLICED_VPS |
                       Zvbi.VBI_SLICED_CAPTION_625 |
                       Zvbi.VBI_SLICED_WSS_625 |
                       Zvbi.VBI_SLICED_VBI_625 )
all_services_525 =   ( Zvbi.VBI_SLICED_CAPTION_525 |
                       Zvbi.VBI_SLICED_2xCAPTION_525 |
                       Zvbi.VBI_SLICED_TELETEXT_BD_525 |
                       Zvbi.VBI_SLICED_VBI_525 )

opt_buf_count = 5
opt_scanning = 0
opt_services = 0

update_services = False
proxy = None # for use in callback

# ---------------------------------------------------------------------------
# Switch channel and frequency (Video 4 Linux #2 API)
#
def SwitchTvChannel(proxy, input_idx, freq):
   result = True

   if (input_idx != -1):
      result = False

      # get current selected input
      try:
         vinp = struct.pack("=i", 0)
         vinp = proxy.device_ioctl(VIDIOC_G_INPUT, vinp)
         prev_inp_idx = struct.unpack("=i", vinp)[0]

         # insert requested channel and norm into the struct
         vinp = struct.pack("=i", input_idx)

         # send channel change request
         try:
            proxy.device_ioctl(VIDIOC_S_INPUT, vinp)
            print("Successfully switched video input from %d to %d" % (prev_inp_idx, input_idx))
            result = True
         except OSError as e:
            print("ioctl VIDIOC_S_INPUT:", e, file=sys.stderr)

      except OSError as e:
         print("ioctl VIDIOC_G_INPUT:", e, file=sys.stderr)

   if (freq != -1):
      result = False

      # query current tuner parameters (including frequency)
      try:
         vfreq = struct.pack("=LLL32x", 0, 0, 0)
         vfreq = proxy.device_ioctl(VIDIOC_G_FREQUENCY, vfreq)
         vtuner, vtype, prev_freq = struct.unpack("=LLL32x", vfreq)

         # send frequency change request
         try:
            vfreq = struct.pack("=LLL32x", vtuner, vtype, freq)
            proxy.device_ioctl(VIDIOC_S_FREQUENCY, vfreq)
            print("Successfully switched frequency: from %d to %d (tuner:%d type:%d)"
                    % (prev_freq, freq, vtuner, vtype), file=sys.stderr)
            result = True
         except OSError as e:
            print("ioctl VIDIOC_S_FREQUENCY(%d):" % freq, e, file=sys.stderr)
      except OSError as e:
         print("ioctl VIDIOC_G_FREQUENCY:", e, file=sys.stderr)

   return result


# ----------------------------------------------------------------------------
# Callback for proxy events
#
def ProxyEventCallback(ev_mask):
   global update_services

   if proxy:
      if (ev_mask & Zvbi.VBI_PROXY_EV_CHN_RECLAIMED):
         print("ProxyEventCallback: token was reclaimed", file=sys.stderr)

         proxy.channel_notify(Zvbi.VBI_PROXY_CHN_TOKEN, 0)

      elif (ev_mask & Zvbi.VBI_PROXY_EV_CHN_GRANTED):
         print("ProxyEventCallback: token granted", file=sys.stderr)

         if ((opt.vinput != -1) or (opt.freq != -1)):
            if (SwitchTvChannel(proxy, opt.vinput, opt.freq)):
               flags = (Zvbi.VBI_PROXY_CHN_TOKEN |
                        Zvbi.VBI_PROXY_CHN_FLUSH)
            else:
               flags = (Zvbi.VBI_PROXY_CHN_RELEASE |
                        Zvbi.VBI_PROXY_CHN_FAIL |
                        Zvbi.VBI_PROXY_CHN_FLUSH)

            if (opt_scanning != 0):
               flags |= Zvbi.VBI_PROXY_CHN_NORM

         else:
            flags = Zvbi.VBI_PROXY_CHN_RELEASE

         proxy.channel_notify(flags, opt_scanning)

      if (ev_mask & Zvbi.VBI_PROXY_EV_CHN_CHANGED):
         lfreq = 0
         # query frequency: tuner, type=V4L2_TUNER_ANALOG_TV, frequency
         buf = struct.pack("=LLL32x", 0, 0, 0)
         try:
            buf = proxy.device_ioctl(VIDIOC_G_FREQUENCY, buf)
            lfreq = struct.unpack("=LLL32x", buf)[2]
            print("Proxy granted video freq: %d" % lfreq, file=sys.stderr)
         except OSError as e:
            print("ProxyEventCallback: VIDIOCGFREQ failed:", e, file=sys.stderr)

         print("ProxyEventCallback: TV channel changed: lfreq", file=sys.stderr)

      if (ev_mask & Zvbi.VBI_PROXY_EV_NORM_CHANGED):
         print("ProxyEventCallback: TV norm changed", file=sys.stderr)
         update_services = True


# ---------------------------------------------------------------------------
# Decode a teletext data line
#
def PrintTeletextData(data, line, slc_id):
   mag    =    0xF
   pkgno  =   0xFF
   tmp1 = Zvbi.unham16p(data)
   if (tmp1 >= 0):
      pkgno = (tmp1 >> 3) & 0x1f
      mag   = tmp1 & 7
      if (mag == 0):
         mag = 8

   if (pkgno != 0):
      data = Zvbi.unpar_str(data, ' ')
      data = re.sub(r'[\x00-\x1F\x7F]', ' ', data.decode('ISO-8859-1'))
      print("line %3d id=%d pkg %X.%03X: '%s'" % (line, slc_id, mag, pkgno, data[2 : 40]))
   else:
      # it's a page header: decode page number and sub-page code
      tmp1 = Zvbi.unham16p(data, 2)
      tmp2 = Zvbi.unham16p(data, 4)
      tmp3 = Zvbi.unham16p(data, 6)
      pageNo = tmp1 | (mag << 8)
      sub    = (tmp2 | (tmp3 << 8)) & 0x3f7f

      data = Zvbi.unpar_str(data, ' ')
      data = re.sub(r'[\x00-\x1F]', ' ', data.decode('ISO-8859-1'))
      print("line %3d id=%d page %03X.%04X: '%s'" % (line, slc_id, pageNo, sub, data[2+8 : 40-8]))


# ---------------------------------------------------------------------------
# Decode a VPS data line
# - bit fields are defined in "VPS Richtlinie 8R2" from August 1995
# - called by the VBI decoder for every received VPS line
#
def PrintVpsData(data):
   VPSOFF = -3

   #cni = ((data[VPSOFF+13] & 0x3) << 10) | ((data[VPSOFF+14] & 0xc0) << 2) |
   #       ((data[VPSOFF+11] & 0xc0)) | (data[VPSOFF+14] & 0x3f)
   #if (cni == 0xDC3):
   #   # special case: "ARD/ZDF Gemeinsames Vormittagsprogramm"
   #   cni = (data[VPSOFF+5] & 0x20) ? 0xDC1 : 0xDC2

   cni = Zvbi.decode_vps_cni(data)

   if ((cni != 0) and (cni != 0xfff)):
      # decode VPS PIL
      mday   =  (data[VPSOFF+11] & 0x3e) >> 1
      month  = ((data[VPSOFF+12] & 0xe0) >> 5) | ((data[VPSOFF+11] & 1) << 3)
      hour   =  (data[VPSOFF+12] & 0x1f)
      minute =  (data[VPSOFF+13] >> 2)

      print("VPS %d.%d. %02d:%02d CNI 0x%04X" % (mday, month, hour, minute, cni))


# ---------------------------------------------------------------------------
# Check stdin for services change requests
# - syntax: ["+"|"-"|"="]keyword, e.g. "+vps-ttx" or "=wss"
#
def read_service_string():
   services = opt_services

   buf = input("")
   if buf != "":
      usage = "Expected command syntax: {+|-|=}{ttx|vps|wss|cc|raw}"
      for match in re.finditer(r' *(\S+)', buf):
         substract = 0
         if (len(match.group(1)) > 1):
            ctrl_str = match.group(1)[0]
            srv_str = match.group(1)[1:]

            if (ctrl_str == "="):
               services = 0
            elif (ctrl_str == "-"):
               substract = 1
            elif (ctrl_str == "+"):
               substract = 0
            else:
               print("Invalid control: '%s' in '%s'\n" % (ctrl_str, buf), usage, file=sys.stderr)

            if ( (srv_str == "ttx") or (srv_str == "teletext") ):
               tmp_services = Zvbi.VBI_SLICED_TELETEXT_B | Zvbi.VBI_SLICED_TELETEXT_BD_525
            elif (srv_str == "vps"):
               tmp_services = Zvbi.VBI_SLICED_VPS
            elif (srv_str == "wss"):
               tmp_services = Zvbi.VBI_SLICED_WSS_625 | Zvbi.VBI_SLICED_WSS_CPR1204
            elif ( (srv_str == "cc") or (srv_str == "caption") ):
               tmp_services = Zvbi.VBI_SLICED_CAPTION_625 | Zvbi.VBI_SLICED_CAPTION_525
            elif (srv_str == "raw"):
               tmp_services = Zvbi.VBI_SLICED_VBI_625 | Zvbi.VBI_SLICED_VBI_525
            else:
               print("Invalid service: '%s'\n" % srv_str, usage, file=sys.stderr)
               break

            if (substract == 0):
               services |= tmp_services
            else:
               services &= ~ tmp_services
         else:
            print("Invalid command (too short): '%s'\n" % buf, usage, file=sys.stderr)
            break

   return services


# ----------------------------------------------------------------------------
# Main loop
#
def main_func():
   global proxy
   global cap
   global update_services
   global opt_services

   #TODO fcntl(STDIN, F_SETFL, O_NONBLOCK)

   if ((opt_services != 0) and (opt_scanning == 0)):
      cur_services = opt_services
   else:
      cur_services = None

   proxy = None
   cap = None
   try:
      if opt.api == 'v4l2':
         cap = Zvbi.Capture.Analog(opt.device, services=cur_services, scanning=opt_scanning,
                            buffers=opt_buf_count, strict=opt.strict, trace=opt.debug_level)
      else:
         proxy = Zvbi.Proxy(opt.device, "proxy-test", trace=opt.debug_level)
         cap = Zvbi.Capture.Analog(opt.device, services=opt_services, scanning=opt_scanning,
                                   buffers=opt_buf_count, strict=opt.strict, trace=opt.debug_level,
                                   proxy=proxy)
         proxy.set_callback(ProxyEventCallback)

      last_line_count = -1

      # switch to the requested channel
      if ( (opt.vinput != -1) or (opt.freq != -1) or
           (opt.chnprio != Zvbi.VBI_CHN_PRIO_INTERACTIVE) ):

         proxy.channel_request(opt.chnprio,
                               request_chn   = (opt.vinput != -1) or (opt.freq != -1),
                               sub_prio      = opt.subprio,
                               exp_duration  = 0,
                               min_duration  = 10)

         if (opt.chnprio != Zvbi.VBI_CHN_PRIO_BACKGROUND):
            SwitchTvChannel(proxy, opt.vinput, opt.freq)

      # initialize services for raw capture
      if ((opt_services & (Zvbi.VBI_SLICED_VBI_625 | Zvbi.VBI_SLICED_VBI_525)) != 0):
         #raw = Zvbi.RawDec(cap.parameters())
         raw = Zvbi.RawDec(cap)
         raw.add_services(all_services_525 | all_services_625, 0)

      update_services = (opt_scanning != 0)

      while True:
         vbi_fd = cap.get_fd()
         if vbi_fd == -1:
            break

         ret, foo1, foo2 = select.select([vbi_fd, sys.stdin.fileno()], [], [])

         if sys.stdin.fileno() in ret:
            new_services = read_service_string()
            if (opt_scanning == 625):
               new_services &= all_services_625
            elif (opt_scanning == 525):
               new_services &= all_services_525

            if (new_services != opt_services):
               print("switching service from 0x%X to 0x%X..." % (opt_services, new_services), file=sys.stderr)
               opt_services = new_services
               update_services = True

         if (update_services):
            try:
               cur_services = cap.update_services(opt_services, reset=True, commit=True, strict=opt.strict)
               if ((cur_services != 0) or (opt_services == 0)):
                  print("...got granted services 0x%X." % cur_services, file=sys.stderr)
            except Zvbi.CaptureError as e:
               print("...failed:", e, file=sys.stderr)

            last_line_count = 0
            update_services = False

         if vbi_fd in ret:
            if ((opt_services & (Zvbi.VBI_SLICED_VBI_625 | Zvbi.VBI_SLICED_VBI_525)) == 0):
               try:
                  sliced_buf = cap.pull_sliced(1000)

                  ttx_lines = 0
                  for (data, slc_id, line) in sliced_buf:
                     if (slc_id & Zvbi.VBI_SLICED_TELETEXT_B):
                        PrintTeletextData(data, line, slc_id)
                        ttx_lines += 1

                     elif (slc_id & Zvbi.VBI_SLICED_VPS):
                        PrintVpsData(data)

                     elif (slc_id & Zvbi.VBI_SLICED_WSS_625):
                        (w0, w1, w2) = struct.unpack("=ccc", data)
                        print("WSS 0x%02X%02X%02X" % (w0, w1, w2))

                     elif (slc_id & (Zvbi.VBI_SLICED_CAPTION_625 | Zvbi.VBI_SLICED_CAPTION_525)):
                        print("CC data 0x%02X,%02X" % (Zvbi.unpar8(data[0]), Zvbi.unpar8(data[1])))

                  if (last_line_count != len(sliced_buf)):
                     last_line_count = len(sliced_buf)
                     print("Receiving frames with %d sliced lines" % last_line_count, file=sys.stderr)

               except Zvbi.CaptureError as e:
                  print("VBI read error:", e, file=sys.stderr)
                  break
               except Zvbi.CaptureTimeout:
                  if (opt.debug_level > 0):
                     print("proxy-test: timeout in VBI read", file=sys.stderr)

            else:
               #res = cap.pull_raw(raw_buf, timestamp, 1000)
               try:
                  raw_buf = cap.read_raw(1000)
                  line_count, sliced = raw.decode(raw_buf)

                  if (last_line_count != line_count):
                     print("Receiving frames with %d slicable lines" % line_count, file=sys.stderr)
                     last_line_count = line_count

                  for (data, slc_id, line) in sliced:
                     if (slc_id & Zvbi.VBI_SLICED_TELETEXT_B):
                        PrintTeletextData(data, line, slc_id)

                     elif (slc_id & Zvbi.VBI_SLICED_VPS):
                        PrintVpsData(data)

                     elif (slc_id & Zvbi.VBI_SLICED_WSS_625):
                        (w0, w1, w2) = struct.unpack("=ccc", data)
                        print("WSS 0x%02X%02X%02X" % (w0, w1, w2))

                     elif (slc_id & (Zvbi.VBI_SLICED_CAPTION_625 | Zvbi.VBI_SLICED_CAPTION_525)):
                        print("CC data 0x%02X,%02X" % (Zvbi.unpar8(data[0]), Zvbi.unpar8(data[1])))

               except Zvbi.CaptureError as e:
                  print("VBI read error:", e, file=sys.stderr)
                  break
               except Zvbi.CaptureTimeout:
                  if (opt.debug_level > 0):
                     print("VBI read timeout", file=sys.stderr)

      del cap

   except Zvbi.ProxyError as e:
      print("Error configuring proxy:", e, file=sys.stderr)
   except Zvbi.CaptureError as e:
      print("Error starting acquisition", e, file=sys.stderr)

   if proxy:
      del proxy

# ---------------------------------------------------------------------------
# Parse command line options
#
def ParseCmdOptions():
   global opt
   global opt_scanning
   global opt_services

   usage = (("%s [ Options ] service ...\n" % sys.argv[0]) +
            "Supported services         : ttx | vps | wss | cc | raw | null\n"+
            "Supported options:\n"+
            "       -dev <path>         : device path\n"+
            "       -api <type>         : v4l API: proxy|v4l2|v4l\n"+
            "       -strict <level>     : service strictness level: 0..2\n"+
            "       -norm PAL|NTSC      : specify video norm as PAL or NTSC\n"+
            "       -vinput <index>     : switch video input source\n"+
            "       -freq <kHz * 16>    : switch TV tuner frequency\n"+
            "       -chnprio <1..3>     : channel switch priority\n"+
            "       -subprio <0..4>     : background scheduling priority\n"+
            "       -debug <level>      : enable debug output: 1=warnings, 2=all\n"+
            "       -help               : this message\n"+
            "You can also type service requests to stdin at runtime:\n"+
            "Format: [\"+\"|\"-\"|\"=\"]<service>, e.g. \"+vps -ttx\" or \"=wss\"\n")

   parser = argparse.ArgumentParser(description="Test of proxy client", usage=usage)
   parser.add_argument("--device", type=str, default="/dev/vbi0", help="Path to VBI capture device (analog only)")
   parser.add_argument("--api", type=str, default="proxy", help="interface to use: proxy|v4l2")
   parser.add_argument("--strict", type=int, default=0, help="service strictness level: 0..2")
   parser.add_argument("--norm", type=str, default="", help="specify video norm as PAL or NTSC")
   parser.add_argument("--vinput", type=int, default=-1, help="switch video input source")
   parser.add_argument("--freq", type=int, default=-1, help="switch TV tuner frequency (unit: kHz*16)")
   parser.add_argument("--chnprio", type=int, default=Zvbi.VBI_CHN_PRIO_INTERACTIVE, help="channel switch priority 1..3")
   parser.add_argument("--subprio", type=int, default=0, help="background scheduling priority 0..4")
   parser.add_argument("--debug", type=int, default=0, dest="debug_level", help="control debug output: 0=none, 1=warnings, 2=all")

   parser.add_argument("srv_lbl", nargs='*', help="services: ttx | vps | wss | cc | raw | null")
   opt = parser.parse_args()

   have_service = False
   opt_services = 0
   for srv in opt.srv_lbl:
      if srv == 'ttx' or srv == 'teletext':
         opt_services |= Zvbi.VBI_SLICED_TELETEXT_B | Zvbi.VBI_SLICED_TELETEXT_BD_525
         have_service = True
      elif srv == 'vps':
         opt_services |= Zvbi.VBI_SLICED_VPS
         have_service = True
      elif srv == 'wss':
         opt_services |= Zvbi.VBI_SLICED_WSS_625 | Zvbi.VBI_SLICED_WSS_CPR1204
         have_service = True
      elif srv == 'cc' or srv == 'caption':
         opt_services |= Zvbi.VBI_SLICED_CAPTION_625 | Zvbi.VBI_SLICED_CAPTION_525
         have_service = True
      elif srv == 'raw':
         opt_services |= Zvbi.VBI_SLICED_VBI_625 | Zvbi.VBI_SLICED_VBI_525
         have_service = True
      elif srv == 'null':
         have_service = True

   if not have_service:
      print("no service given - Must specify at least one service (may be 'null')", file=sys.stderr)
      sys.exit(1)

   if not (opt.api == "proxy" or opt.api == "v4l2"):
      print("Unknown API:", opt.api, "\n", usage, file=sys.stderr)
      exit(1)

   if opt.norm == "PAL":
      opt_scanning = 625
   elif opt.norm == "NTSC":
      opt_scanning = 525
   elif not opt.norm == "":
      print("Unknown video norm:", opt.norm, "\n", usage, file=sys.stderr)
      exit(1)

   if opt.strict < 0 or opt.strict > 2:
      print("Invalid strictness level:", opt.strict, "\n", usage, file=sys.stderr)
      exit(1)

   if "dvb" in opt.device:
      print("WARNING: DVB devices are not supported by proxy", file=sys.stderr)


# ---------------------------------------------------------------------------
# main

try:
   ParseCmdOptions()
   main_func()
except KeyboardInterrupt:
   pass
